/* Copyright (C) 1999-2000 Hans Petter K. Jansson
 *
 * This library is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
 *
 * You can contact the library's author by sending e-mail to <hpj@styx.net>.
 */

#include <stdio.h>
#include <signal.h>
#include <unistd.h>

#include "config.h"
#include "flux.h"

#define SERVER_PORT 2140

PROXY *p;
SOCK *s_peer, *s_ppp;


char *cmd_args[] =
{
   "", "notty", "passive", "local", "noauth", 0
};

int _interrupted = 0;


void handle_interrupt(int signal)
{
  _interrupted = 1;
}


int peer_read(SOCK *s, void *priv)
{
  unsigned char buf[4096];
  int len;
  
  len = sock_get_max_read(s_peer);
  if (len > 4096) len = 4096;
  if (len > sock_get_max_write(s_ppp)) len = sock_get_max_write(s_ppp);

  if (len)
  {
    sock_read(s_peer, buf, len);
    sock_write(s_ppp, buf, len);
  }
  
  return(0);
}


int ppp_read(SOCK *s, void *priv)
{
  unsigned char buf[4096];
  int len;
  
  len = sock_get_max_read(s_ppp);
  if (len > 4096) len = 4096;
  if (len > sock_get_max_write(s_peer)) len = sock_get_max_write(s_peer);

  if (len)
  {
    sock_read(s_ppp, buf, len);
    sock_write(s_peer, buf, len);
  }
  
  return(0);
}


int disconnect(SOCK *s, void *priv)
{
  if ((int) priv == 1)
    fprintf(stderr, "Peer disconnected!\n");
  else
    fprintf(stderr, "PPP interface disconnected!\n");

  proxy_break(p);
  return(0);
}


void handle_request(SOCK *s)
{
  signal(SIGCLD, SIG_IGN);
  signal(SIGINT, handle_interrupt);
  signal(SIGHUP, handle_interrupt);

  s_peer = s;
  p = proxy_new();

  if (!(s_ppp = proc_exec_sock("/usr/sbin/pppd", cmd_args, 0)))
  {
    sock_close(s_peer, 0);
    return;
  }
  
  /* Add to proxy */

  proxy_add_sock_read(p, s_peer, 0, peer_read);
  proxy_add_sock_read(p, s_ppp, 0, ppp_read);
  proxy_add_sock_write(p, s_peer, 0, ppp_read);
  proxy_add_sock_write(p, s_ppp, 0, peer_read);
  proxy_add_sock_close(p, s_peer, (void *) 1, disconnect);
  proxy_add_sock_close(p, s_ppp, 0, disconnect);

  /* Read/write */

  proxy_loop(p);
  
  /* Exit */

  proxy_del_sock(p, s_peer);
  proxy_del_sock(p, s_ppp);
  proxy_del(p);

  sock_del(s_peer);
  sock_del(s_ppp);
}


int main()
{
  CIPHER *cip;
  SOCK *s_parent, *s_child;


  signal(SIGCLD, SIG_IGN);  /* Dispel undead (zombie daemon children) */
  signal(SIGINT, handle_interrupt);

  /* Generate a secret key */
  
  printf("Generating server key..."); fflush(stdout);

  cip = cipher_new(CIPHER_NONE, CIPHER_ELGAMAL);
  cip->asym_key_decrypt = elgamal_gen_secret(768);
  cip->asym_key_sign = cip->asym_key_decrypt;
  cip->asym_key_encrypt = elgamal_gen_public(cip->asym_key_decrypt);
  cip->asym_key_verify = cip->asym_key_encrypt;
  
  printf("\n"); fflush(stdout);

  /* Open socket and assign key */

  s_parent = sock_open(SERVER_PORT, SOCK_ACCEPT | SOCK_LOG | SOCK_DEBUG);
  if (!s_parent) { fprintf(stderr, "Server: Unable to open socket.\n"); return(0); }

  sock_set_cipher(s_parent, cip);

  printf("Server: Listening."); fflush(stdout);
  for (;;)
  {
    s_child = sock_wait(0, 0, s_parent, 0);

    if (s_child)
    {
      printf("\nServer: Got request on port %d.\n", s_child->parent->port_bound);
      if (!fork()) { sock_close(s_parent, 0); handle_request(s_child); }
      sock_close(s_child, 0);
    }

    if (_interrupted) break;
  }

  printf("\nServer: Terminated by CTRL-C.\n");
  sock_close(s_parent, 0);
  return(0);
}
