/* Copyright (C) 1999-2000 Hans Petter K. Jansson
 *
 * This library is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
 *
 * You can contact the library's author by sending e-mail to <hpj@styx.net>.
 */

#include "config.h"
#include "types.h"
#include "bmem.h"
#include "bt.h"
#include "bt_priv.h"
#include <stdlib.h>


#define bt_node_is_red(btn) ((btn)->is_red)
#define bt_node_set_red(btn, b) ((btn)->is_red = (b))


#define bt_node_rotate(key, x, y)                                      \
{                                                                      \
  BTN *__c, *__gc;                                                     \
                                                                       \
  __c = (key < y->key) ? y->left : y->right;                           \
                                                                       \
  if (key < __c->key)                                                  \
  { __gc = __c->left; __c->left = __gc->right; __gc->right = __c; }    \
  else                                                                 \
  { __gc = __c->right; __c->right = __gc->left; __gc->left = __c; }    \
                                                                       \
  if (key < y->key) y->left = __gc; else y->right = __gc;              \
                                                                       \
  x = __gc;                                                            \
}


#define bt_node_split(x, p, g, gg, key)                                \
{                                                                      \
  bt_node_set_red(x, TRUE);                                            \
  bt_node_set_red(x->left, FALSE);                                     \
  bt_node_set_red(x->right, FALSE);                                    \
                                                                       \
  if (bt_node_is_red(p))                                               \
  {                                                                    \
    bt_node_set_red(g, TRUE);                                          \
    if ((key < g->key) != (key < p->key)) bt_node_rotate(key, p, g);   \
    bt_node_rotate(key, x, gg);                                        \
    bt_node_set_red(x, FALSE);                                         \
  }                                                                    \
}


void bt_insert(BT *bt, void *key, void *data)
{
  BTN *x, *p, *g, *gg;  /* x, parent, grandparent, greatgrandparent */

  /* Traverse the tree, splitting 4-nodes */
  
  for (x = p = g = gg = &bt->head; x != &bt->z; )  /* gg: superfluous */
  {
    gg = g; g = p; p = x;
    
    x = (key < x->key) ? x->left : x->right;
    if (bt_node_is_red(x->left) && bt_node_is_red(x->right))
      bt_node_split(x, p, g, gg, key);
  }

  /* Alloc and init new node */

  if (bt->bmem) x = bmem_block_new(bt->bmem);
  else x = malloc(sizeof(BTN));

  bt_node_set_red(x, TRUE);
  x->key = key; x->data = data;
  x->left = x->right = &bt->z;

  /* Link it in */
  
  if (key < p->key) p->left = x;
  else p->right = x;
  
  /* Restructure */
  
  bt_node_split(x, p, g, gg, key);
  bt_node_set_red(bt->head.right, FALSE);
}


void bt_move(BT *btd, BT *bts, void *key)
{
  BTN *c, *x, *t, *p, *g, *gg;

  /* --- Removal --- */
  
  p = &bts->head;
  x = bts->head.right;
  bts->z.key = key;
  
  while (key != x->key)
  {
    p = x;
    x = (key < x->key) ? x->left : x->right;
  }

  if (x == &bts->z) return;  /* No such key; let's not remove the sentinel node */

  t = x;
  if (t->right == &bts->z) x = x->left;
  else if (t->right->left == &bts->z)
  {
    x = x->right;
    x->left = t->left;
  }
  else
  {
    c = x->right;
    while (c->left->left != &bts->z) c = c->left;
    x = c->left;
    c->left = x->right;
    x->left = t->left;
    x->right = t->right;
  }
  
  /* t = detached node */

  if (key < p->key) p->left = x;
  else p->right = x;

  /* --- Insertion --- */
  
  /* Traverse the tree, splitting 4-nodes */
  
  for (x = p = g = gg = &btd->head; x != &btd->z; )  /* gg: superfluous */
  {
    gg = g; g = p; p = x;
    
    x = (key < x->key) ? x->left : x->right;
    if (bt_node_is_red(x->left) && bt_node_is_red(x->right))
      bt_node_split(x, p, g, gg, key);
  }

  /* Alloc and init new node */

  x = t;
  bt_node_set_red(x, TRUE);
  x->left = x->right = &btd->z;

  /* Link it in */
  
  if (key < p->key) p->left = x;
  else p->right = x;
  
  /* Restructure */
  
  bt_node_split(x, p, g, gg, key);
  bt_node_set_red(btd->head.right, FALSE);
}
