/* Copyright (C) 1999-2000 Hans Petter K. Jansson
 *
 * This library is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
 *
 * You can contact the library's author by sending e-mail to <hpj@styx.net>.
 */

#include "config.h"
#include "types.h"
#include "log.h"
#include "cipher.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>


CIPHER *cipher_new(CIPHER_TYPE sym, CIPHER_TYPE asym)
{
  CIPHER *cip;
  
  cip = malloc(sizeof(*cip));
  memset(cip, 0, sizeof(*cip));
  
  cip->sym_type = sym;
  cip->asym_type = asym;

  return(cip);
}


void cipher_del(CIPHER *cip)
{
  if (cip->sym_type == CIPHER_TWOFISH)
  {
    if (cip->sym_key_encrypt) twofish_del(cip->sym_key_encrypt);
    if (cip->sym_key_decrypt && cip->sym_key_decrypt != cip->sym_key_encrypt)
      twofish_del(cip->sym_key_decrypt);
  }

  if (cip->asym_type == CIPHER_ELGAMAL)
  {
    if (cip->asym_key_encrypt) elgamal_del_public(cip->asym_key_encrypt);
    if (cip->asym_key_verify && cip->asym_key_verify != cip->asym_key_encrypt)
      elgamal_del_public(cip->asym_key_verify);
    
    if (cip->asym_key_decrypt) elgamal_del_secret(cip->asym_key_decrypt);
    if (cip->asym_key_sign && cip->asym_key_sign != cip->asym_key_decrypt)
      elgamal_del_secret(cip->asym_key_sign);
  }

  memset(cip, 0, sizeof(*cip));
  free(cip);
}


CIPHER *cipher_dup(CIPHER *old)
{
  CIPHER *new;
  
  new = cipher_new(old->sym_type, old->asym_type);
  
  if (old->sym_type == CIPHER_TWOFISH)
  {
    if (old->sym_key_encrypt) new->sym_key_encrypt = twofish_dup(old->sym_key_encrypt);
    if (old->sym_key_decrypt && old->sym_key_decrypt != old->sym_key_encrypt)
      new->sym_key_decrypt = twofish_dup(old->sym_key_decrypt);
    else
      new->sym_key_decrypt = new->sym_key_encrypt;
  }

  if (old->asym_type == CIPHER_ELGAMAL)
  {
    if (old->asym_key_encrypt) new->asym_key_encrypt = elgamal_dup_public(old->asym_key_encrypt);
    if (old->asym_key_verify && old->asym_key_verify != old->asym_key_encrypt)
      new->asym_key_verify = elgamal_dup_public(old->asym_key_verify);
    else
      new->asym_key_verify = new->asym_key_encrypt;

    if (old->asym_key_decrypt) new->asym_key_decrypt = elgamal_dup_secret(old->asym_key_decrypt);
    if (old->asym_key_sign && old->asym_key_sign != old->asym_key_decrypt)
      new->asym_key_sign = elgamal_dup_secret(old->asym_key_sign);
    else
      new->asym_key_sign = new->asym_key_decrypt;
  }

  return(new);
}


void cipher_gen(CIPHER *cip)
{
  if (cip->sym_type != CIPHER_NONE)
  {
    /* We only know Twofish right now */
    
    /* Free any previous keys */
    
    if (cip->sym_key_encrypt) twofish_del(cip->sym_key_encrypt);
    if (cip->sym_key_decrypt && cip->sym_key_decrypt != cip->sym_key_encrypt)
      twofish_del(cip->sym_key_decrypt);

    /* At this time, cip->sym_type can be CIPHER_ANY */

    cip->sym_type = CIPHER_TWOFISH;

    cip->sym_key_encrypt = twofish_gen(128);
    cip->sym_key_decrypt = twofish_dup(cip->sym_key_encrypt);
  }

  if (cip->asym_type != CIPHER_NONE)
  {
    /* We only know ElGamal right now */
    
    /* Free any previous keys */

    if (cip->asym_key_encrypt) elgamal_del_public(cip->asym_key_encrypt);
    if (cip->asym_key_verify && cip->asym_key_verify != cip->asym_key_encrypt)
      elgamal_del_public(cip->asym_key_verify);
    
    if (cip->asym_key_decrypt) elgamal_del_secret(cip->asym_key_decrypt);
    if (cip->asym_key_sign && cip->asym_key_sign != cip->asym_key_decrypt)
      elgamal_del_secret(cip->asym_key_sign);
    
    /* At this time, cip->asym_type can be CIPHER_ANY */
    
    cip->asym_type = CIPHER_ELGAMAL;
    
    cip->asym_key_decrypt = cip->asym_key_sign = elgamal_gen_secret(1984);
    cip->asym_key_encrypt = cip->asym_key_verify =
      elgamal_gen_public(cip->asym_key_decrypt);
  }
}


int cipher_has_keys(CIPHER *cip)
{
  if ((cip->sym_type == CIPHER_TWOFISH) && (!cip->sym_key_encrypt ||
      cip->sym_key_decrypt))
    return(0);

  if ((cip->asym_type == CIPHER_ELGAMAL) && (!cip->asym_key_encrypt ||
      !cip->asym_key_decrypt || !cip->asym_key_sign || !cip->asym_key_verify))
    return(0);

  return(1);
}


void cipher_export_asym(CIPHER *cip, byte **out, unsigned int *len)
{
  byte *buf;
  
  if (cip->asym_type == CIPHER_ELGAMAL)
  {
    elgamal_export_public(cip->asym_key_encrypt, out, len);
    
    /* Insert algorithm and total length */
    
    buf = malloc(*len + 3);
    buf[0] = CIPHER_ELGAMAL;
    buf[1] = *len >> 8;
    buf[2] = *len & 0xff;
    memcpy(buf + 3, *out, *len);
    
    free(*out);
    *out = buf;
    *len = *len + 3;

#if 0
    log_put_opt(LOG_DEBUG, 0, "[Cipher] Exported asymmmetric - length is %d (%d).",
                *len, (buf[1] << 8) + buf[2]);
#endif
  }
  else
  {
    *out = 0;
    *len = 0;
  }
}


void cipher_export_sym(CIPHER *cip, byte **out, unsigned int *len)
{
  byte *buf;
  byte *iv;
  unsigned int iv_len;
  
  if (cip->sym_type == CIPHER_TWOFISH)
  {
    twofish_export(cip->sym_key_encrypt, out, len);
    twofish_export_iv(cip->sym_key_encrypt, &iv, &iv_len);

    /* Insert algorithm, key length, iv length and iv */

    buf = malloc(*len + iv_len + 3);
    buf[0] = CIPHER_TWOFISH;
    buf[1] = *len;
    buf[2] = (byte) iv_len;
    memcpy(buf + 3, *out, *len);
    memcpy(buf + 3 + *len, iv, iv_len);

    free(iv);
    free(*out);
    *out = buf;
    *len = *len + iv_len + 3;
  }
  else
  {
    *out = 0;
    *len = 0;
  }
}


int cipher_import_asym(CIPHER *cip, byte *in)
{
  void *key;
  
  /* TODO: Free old (residue) keys */
  
  if (in[0] == CIPHER_ELGAMAL)
  {
    key = elgamal_import_public(in + 3);
    cip->asym_type = CIPHER_ELGAMAL;
  }
  else
  {
    /* Unknown algorithm */

    return(-1);
  }

  cip->asym_key_encrypt = cip->asym_key_verify = key;
  return(0);
}


int cipher_import_sym(CIPHER *cip, byte *in)
{
  void *key;
  
  /* TODO: Free old (residue) keys */
  
  if (in[0] == CIPHER_TWOFISH)
  {
    key = twofish_import(in + 3, in[1]);
    twofish_import_iv(key, in + 3 + in[1]);
    cip->sym_type = CIPHER_TWOFISH;
  }
  else
  {
    /* Unknown algorithm */
    
    return(-1);
  }
  
  cip->sym_key_encrypt = cip->sym_key_decrypt = key;
  return(0);
}


void cipher_permute_in(CIPHER *cip)
{
  if (cip->sym_type == CIPHER_TWOFISH)
  {
    /* If the en/decryption keys are the same, make a duplicate before
     * permuting one. */

    if (cip->sym_key_encrypt == cip->sym_key_decrypt)
      cip->sym_key_decrypt = twofish_dup(cip->sym_key_decrypt);

    twofish_permute_iv(cip->sym_key_decrypt);
  }
  else
  {
    /* No cipher, or we don't know how to permute */
    
    return;
  }
}


void cipher_permute_out(CIPHER *cip)
{
  if (cip->sym_type == CIPHER_TWOFISH)
  {
    /* If the en/decryption keys are the same, make a duplicate before
     * permuting one. */

    if (cip->sym_key_encrypt == cip->sym_key_decrypt)
      cip->sym_key_encrypt = twofish_dup(cip->sym_key_encrypt);

    twofish_permute_iv(cip->sym_key_encrypt);
  }
  else
  {
    /* No cipher, or we don't know how to permute */
    
    return;
  }
}
