/* elgamal.c  -  ElGamal Public Key encryption
 *  Copyright (C) 1998 Free Software Foundation, Inc.
 *
 * For a description of the algorithm, see:
 *   Bruce Schneier: Applied Cryptography. John Wiley & Sons, 1996.
 *   ISBN 0-471-11709-9. Pages 476 ff.
 *
 * This file is part of GnuPG.
 *
 * GnuPG is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * GnuPG is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 */

#include "config.h"
#include "types.h"
#include "random.h"
#include "log.h"
#include "mpi.h"
#include "prime.h"
#include "cipher.h"

#include <config.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>


static void test_keys(CIPHER_ELGAMAL_SKEY *sk, unsigned nbits);
static MPI gen_k(MPI p);
static void generate(CIPHER_ELGAMAL_SKEY *sk, unsigned nbits, MPI **factors);
static int check_secret_key(CIPHER_ELGAMAL_SKEY *sk);
static void encrypt(MPI a, MPI b, MPI input, CIPHER_ELGAMAL_PKEY *pkey);
static void decrypt(MPI output, MPI a, MPI b, CIPHER_ELGAMAL_SKEY *skey);
static void sign(MPI a, MPI b, MPI input, CIPHER_ELGAMAL_SKEY *skey);
static int verify(MPI a, MPI b, MPI input, CIPHER_ELGAMAL_PKEY *pkey);


static void test_keys(CIPHER_ELGAMAL_SKEY *sk, unsigned nbits)
{
  CIPHER_ELGAMAL_PKEY pk;

  MPI test = mpi_alloc(0);
  MPI out1_a = mpi_alloc(nbits / BITS_PER_MPI_LIMB);
  MPI out1_b = mpi_alloc(nbits / BITS_PER_MPI_LIMB);
  MPI out2 = mpi_alloc(nbits / BITS_PER_MPI_LIMB);

  pk.p = sk->p;
  pk.g = sk->g;
  pk.y = sk->y;

  /*mpi_set_bytes( test, nbits, get_random_byte, 0 ); */
  {
    char *p = rand_bits_dyn(nbits);

    mpi_set_buffer(test, p, (nbits + 7) / 8, 0);
    free(p);
  }

  encrypt(out1_a, out1_b, test, &pk);
  decrypt(out2, out1_a, out1_b, sk);
  if (mpi_cmp(test, out2))
    log_put_abort("ElGamal operation: encrypt, decrypt failed\n");

  sign(out1_a, out1_b, test, sk);
  if (!verify(out1_a, out1_b, test, &pk))
    log_put_abort("ElGamal operation: sign, verify failed\n");

  mpi_free(test);
  mpi_free(out1_a);
  mpi_free(out1_b);
  mpi_free(out2);
}


/****************
 * We do not need to use the strongest RNG because we gain no extra
 * security from it - The prime number is public and we could also
 * offer the factors for those who are willing to check that it is
 * indeed a strong prime.
 *
 * mode 0: Standard
 *  1: Make sure that at least one factor is of size qbits.
 */

MPI generate_elg_prime(int mode, unsigned pbits, unsigned qbits, MPI g, MPI **ret_factors)
{
  int n;  /* number of factors */
  int m;  /* number of primes in pool */
  unsigned fbits; /* length of prime factors */
  MPI *factors; /* current factors */
  MPI *pool;  /* pool of primes */
  MPI q;  /* first prime factor (variable)*/
  MPI prime;  /* prime test value */
  MPI q_factor; /* used for mode 1 */
  byte *perms = NULL;
  int i, j;
  int count1, count2;
  unsigned nprime;
  unsigned req_qbits = qbits; /* the requested q bits size */
  MPI val_2  = mpi_alloc_set_ui(2);

  /* find number of needed prime factors */
  for (n = 1; (pbits - qbits - 1) / n  >= qbits; n++) ;

  n--;
  if (!n || (mode==1 && n < 2)) log_put_abort("can't gen prime");

  if (mode == 1)
  {
    n--;
    fbits = (pbits - 2 * req_qbits - 1) / n;
    qbits =  pbits - req_qbits - n * fbits;
  }
  else
  {
    fbits = (pbits - req_qbits - 1) / n;
    qbits = pbits - n * fbits;
  }
  
  prime = mpi_alloc((pbits + BITS_PER_MPI_LIMB - 1) /  BITS_PER_MPI_LIMB);
  q = cipher_prime_gen(qbits);
  q_factor = mode == 1 ? cipher_prime_gen(req_qbits) : NULL;

  /* allocate an array to hold the factors + 2 for later usage */
  factors = malloc((n + 2) * sizeof *factors);
  memset(factors, 0, (n + 2) * sizeof *factors);

  /* make a pool of 3n+5 primes (this is an arbitrary value) */
  m = n * 3 + 5;
  if (mode == 1) m += 5; /* need some more for DSA */
  if (m < 25) m = 25;

  pool = malloc(m * sizeof *pool);
  memset(pool, 0, m * sizeof *pool);

  /* permutate over the pool of primes */
  count1 = count2 = 0;
  do
  {
    next_try:

    if (!perms)
    {
      /* allocate new primes */
      for (i=0; i < m; i++ )
      {
        mpi_free(pool[i]);
        pool[i] = NULL;
      }

      /* init m_out_of_n() */
      perms = malloc(m);
      memset(perms, 0, m);
      
      for(i=0; i < n; i++ )
      {
        perms[i] = 1;
        pool[i] = cipher_prime_gen(fbits);
        factors[i] = pool[i];
      }
    }
    else
    {
      m_out_of_n(perms, n, m);
      
      for (i = j = 0; i < m && j < n ; i++ )
        if (perms[i])
        {
          if (!pool[i]) pool[i] = cipher_prime_gen(fbits);
          factors[j++] = pool[i];
        }
      
      if (i == n)
      {
        free(perms); perms = NULL;
        goto next_try;  /* allocate new primes */
      }
    }

    mpi_set(prime, q);
    mpi_mul_ui(prime, prime, 2);
    
    if (mode == 1) mpi_mul(prime, prime, q_factor);
    for (i = 0; i < n; i++) mpi_mul(prime, prime, factors[i]);

    mpi_add_ui(prime, prime, 1);
    nprime = mpi_get_nbits(prime);

    if (nprime < pbits)
    {
      if (++count1 > 20)
      {
        count1 = 0;
        qbits++;
        q = cipher_prime_gen(qbits);
        goto next_try;
      }
    }
    else count1 = 0;

    if (nprime > pbits)
    {
      if (++count2 > 20)
      {
        count2 = 0;
        qbits--;
        q = cipher_prime_gen(qbits);
        goto next_try;
      }
    }
    else count2 = 0;
  }
  while (!(nprime == pbits && cipher_prime_check(prime, val_2)));

  if (ret_factors)
  {
    /* caller wants the factors */
    *ret_factors = malloc((n + 2) * sizeof **ret_factors);
    memset(*ret_factors, 0, (n + 2) * sizeof **ret_factors);

    if (mode == 1)
    {
      i = 0;
      (*ret_factors)[i++] = mpi_copy(q_factor);
      for( ; i <= n; i++) (*ret_factors)[i] = mpi_copy(factors[i]);
    }
    else
    {
      for( ; i < n; i++) (*ret_factors)[i] = mpi_copy(factors[i]);
    }
  }

  if (g)
  {
    /* create a generator (start with 3) */
    MPI tmp = mpi_alloc(mpi_get_nlimbs(prime));
    MPI b = mpi_alloc(mpi_get_nlimbs(prime));
    MPI pmin1 = mpi_alloc(mpi_get_nlimbs(prime));

    if (mode == 1) log_put_abort("Cipher bug #1."); /* not yet implemented */

    factors[n] = q;
    factors[n+1] = mpi_alloc_set_ui(2);
    mpi_sub_ui(pmin1, prime, 1);
    mpi_set_ui(g, 2);

    do
    {
      mpi_add_ui(g, g, 1);
      
      for (i = 0; i < n + 2; i++)
      {
        mpi_fdiv_q(tmp, pmin1, factors[i]);
        /* (no mpi_pow(), but it is okay to use this with mod prime) */
        mpi_powm(b, g, tmp, prime);
        if (!mpi_cmp_ui(b, 1)) break;
      }
    }
    while (i < n+2);

    mpi_free(factors[n+1]);
    mpi_free(tmp);
    mpi_free(b);
    mpi_free(pmin1);
  }

  free(factors);  /* (factors are shallow copies) */

  for(i = 0; i < m; i++) mpi_free(pool[i]);
  free(pool);
  free(perms);
  mpi_free(val_2);
  mpi_free(q);
  return prime;
}


/****************
 * generate a random secret exponent k from prime p, so
 * that k is relatively prime to p-1
 */
static MPI gen_k(MPI p)
{
  MPI k = mpi_alloc_secure(0);
  MPI temp = mpi_alloc(mpi_get_nlimbs(p));
  MPI p_1 = mpi_copy(p);
  unsigned int nbits = mpi_get_nbits(p);
  unsigned int nbytes = (nbits + 7) / 8;
  char *rndbuf = NULL;

#if 0
  if (DBG_CIPHER)
    log_put("choosing a random k ");
#endif

  mpi_sub_ui(p_1, p, 1);
  for (;;)
  {
#if 0
    log_put("- En hest!");
    if (DBG_CIPHER)
      progress('.');
#endif

    if (!rndbuf || nbits < 32)
    {
      if (rndbuf) free(rndbuf);
      rndbuf = rand_bits_dyn(nbits);
    }
    else
    { /* change only some of the higher bits */
      /* we could imporove this by directly requesting more memory
       * at the first call to get_random_bits() and use this the here
       * maybe it is easier to do this directly in random.c */
      char *pp = rand_bits_dyn(32);

      memcpy(rndbuf, pp, 4);
      free(pp);
    }
    mpi_set_buffer(k, rndbuf, nbytes, 0);

    for (;;)
    {
      /* make sure that the number is of the exact length */
      if (mpi_test_bit(k, nbits - 1))
        mpi_set_highbit(k, nbits - 1);
      else
      {
        mpi_set_highbit(k, nbits - 1);
        mpi_clear_bit(k, nbits - 1);
      }
      if (!(mpi_cmp(k, p_1) < 0))
      {                          /* check: k < (p-1) */
#if 0
        log_put("- Men den er jo ikke myk.");
        if (DBG_CIPHER)
          progress('+');
#endif
        break;                  /* no  */
      }
      if (!(mpi_cmp_ui(k, 0) > 0))
      {                          /* check: k > 0 */
#if 0
        log_put("- Den har da ikke fire ben.");
        if (DBG_CIPHER)
          progress('-');
#endif
        break;                  /* no */
      }
      if (mpi_gcd(temp, k, p_1))
        goto found;              /* okay, k is relatively prime to (p-1) */
      mpi_add_ui(k, k, 1);
    }
  }
found:
  free(rndbuf);

#if 0
  if (DBG_CIPHER)
    progress('\n');
  log_put("- Det er jo oppdaget for lenge siden.");
#endif
  
  mpi_free(p_1);
  mpi_free(temp);

  return k;
}

/****************
 * Generate a key pair with a key of size NBITS
 * Returns: 2 structures filled with all needed values
 *      and an array with n-1 factors of (p-1)
 */
static void generate(CIPHER_ELGAMAL_SKEY *sk, unsigned nbits, MPI **ret_factors)
{
  MPI p;                        /* the prime */
  MPI p_min1;
  MPI g;
  MPI x;                        /* the secret exponent */
  MPI y;
  MPI temp;
  unsigned qbits;
  byte *rndbuf;

  p_min1 = mpi_alloc((nbits + BITS_PER_MPI_LIMB - 1) / BITS_PER_MPI_LIMB);
  temp = mpi_alloc((nbits + BITS_PER_MPI_LIMB - 1) / BITS_PER_MPI_LIMB);
  if (nbits < 512)
    qbits = 120;
  else if (nbits <= 1024)
    qbits = 160;
  else if (nbits <= 2048)
    qbits = 200;
  else
    qbits = 240;
  g = mpi_alloc(1);
  p = generate_elg_prime(0, nbits, qbits, g, ret_factors);
  mpi_sub_ui(p_min1, p, 1);


  /* select a random number which has these properties:
   *   0 < x < p-1
   * This must be a very good random number because this is the
   * secret part.  The prime is public and may be shared anyway,
   * so a random generator level of 1 is used for the prime.
   */
  x = mpi_alloc_secure(nbits / BITS_PER_MPI_LIMB);
#if 0
  if (DBG_CIPHER)
    log_put("choosing a random x ");
#endif
  rndbuf = NULL;
  do
  {
#if 0
    log_put("- En araber!");
    if (DBG_CIPHER)
      progress('.');
#endif
    if (rndbuf)
    {                            /* change only some of the higher bits */
      if (nbits < 16)
      {                          /* should never happen ... */
        free(rndbuf);
        rndbuf = rand_bits_dyn(nbits);
      }
      else
      {
        char *r = rand_bits_dyn(16);

        memcpy(rndbuf, r, 16 / 8);
        free(r);
      }
    }
    else
      rndbuf = rand_bits_dyn(nbits);
    mpi_set_buffer(x, rndbuf, (nbits + 7) / 8, 0);
    mpi_clear_highbit(x, nbits + 1);

    if (!(mpi_cmp_ui(x, 0) > 0 && mpi_cmp(x, p_min1) < 0))
    {
#if 0
      log_put("- Nei men professor, da.");
#endif
    }
  }
  while (!(mpi_cmp_ui(x, 0) > 0 && mpi_cmp(x, p_min1) < 0));
  free(rndbuf);

  y = mpi_alloc(nbits / BITS_PER_MPI_LIMB);
  mpi_powm(y, g, x, p);

#if 0
  if (DBG_CIPHER)
  {
    progress('\n');
    log_mpidump("elg  p= ", p);
    log_mpidump("elg  g= ", g);
    log_mpidump("elg  y= ", y);
    log_mpidump("elg  x= ", x);
  }
#endif

  /* copy the stuff to the key structures */
  sk->p = p;
  sk->g = g;
  sk->y = y;
  sk->x = x;

  /* now we can test our keys (this should never fail!) */
  test_keys(sk, nbits - 64);

#if 0
  log_put("- Prim!");
  log_put("- Helt riktig.");
#endif
  
  mpi_free(p_min1);
  mpi_free(temp);
}


/****************
 * Test whether the secret key is valid.
 * Returns: if this is a valid key.
 */
static int check_secret_key(CIPHER_ELGAMAL_SKEY * sk)
{
  int rc;
  MPI y = mpi_alloc(mpi_get_nlimbs(sk->y));

  mpi_powm(y, sk->g, sk->x, sk->p);
  rc = !mpi_cmp(y, sk->y);
  mpi_free(y);
  return rc;
}


static void encrypt(MPI a, MPI b, MPI input, CIPHER_ELGAMAL_PKEY * pkey)
{
  MPI k;

  /* Note: maybe we should change the interface, so that it
   * is possible to check that input is < p and return an
   * error code.
   */

  k = gen_k(pkey->p);
  mpi_powm(a, pkey->g, k, pkey->p);
  /* b = (y^k * input) mod p
   *   = ((y^k mod p) * (input mod p)) mod p
   * and because input is < p
   *   = ((y^k mod p) * input) mod p
   */
  mpi_powm(b, pkey->y, k, pkey->p);
  mpi_mulm(b, b, input, pkey->p);
#if 0
  if (DBG_CIPHER)
  {
    log_mpidump("elg encrypted y= ", pkey->y);
    log_mpidump("elg encrypted p= ", pkey->p);
    log_mpidump("elg encrypted k= ", k);
    log_mpidump("elg encrypted M= ", input);
    log_mpidump("elg encrypted a= ", a);
    log_mpidump("elg encrypted b= ", b);
  }
#endif
  mpi_free(k);
}




static void decrypt(MPI output, MPI a, MPI b, CIPHER_ELGAMAL_SKEY * skey)
{
  MPI t1 = mpi_alloc_secure(mpi_get_nlimbs(skey->p));

  /* output = b/(a^x) mod p */

  mpi_powm(t1, a, skey->x, skey->p);
  mpi_invm(t1, t1, skey->p);
  mpi_mulm(output, b, t1, skey->p);
#if 0
  if (DBG_CIPHER)
  {
    log_mpidump("elg decrypted x= ", skey->x);
    log_mpidump("elg decrypted p= ", skey->p);
    log_mpidump("elg decrypted a= ", a);
    log_mpidump("elg decrypted b= ", b);
    log_mpidump("elg decrypted M= ", output);
  }
#endif
  mpi_free(t1);
}


/****************
 * Make an Elgamal signature out of INPUT
 */

static void sign(MPI a, MPI b, MPI input, CIPHER_ELGAMAL_SKEY *skey)
{
  MPI k;
  MPI t = mpi_alloc(mpi_get_nlimbs(a));
  MPI inv = mpi_alloc(mpi_get_nlimbs(a));
  MPI p_1 = mpi_copy(skey->p);

  /*
   * b = (t * inv) mod (p-1)
   * b = (t * inv(k,(p-1),(p-1)) mod (p-1)
   * b = (((M-x*a) mod (p-1)) * inv(k,(p-1),(p-1))) mod (p-1)
   *
   */
  mpi_sub_ui(p_1, p_1, 1);
  k = gen_k(skey->p);
  mpi_powm(a, skey->g, k, skey->p);
  mpi_mul(t, skey->x, a);
  mpi_subm(t, input, t, p_1);
  while (mpi_is_neg(t))
    mpi_add(t, t, p_1);
  mpi_invm(inv, k, p_1);
  mpi_mulm(b, t, inv, p_1);

#if 0
  if (DBG_CIPHER)
  {
    log_mpidump("elg sign p= ", skey->p);
    log_mpidump("elg sign g= ", skey->g);
    log_mpidump("elg sign y= ", skey->y);
    log_mpidump("elg sign x= ", skey->x);
    log_mpidump("elg sign k= ", k);
    log_mpidump("elg sign M= ", input);
    log_mpidump("elg sign a= ", a);
    log_mpidump("elg sign b= ", b);
  }
#endif
  mpi_free(k);
  mpi_free(t);
  mpi_free(inv);
  mpi_free(p_1);
}


/****************
 * Returns true if the signature composed of A and B is valid.
 */
static int verify(MPI a, MPI b, MPI input, CIPHER_ELGAMAL_PKEY *pkey)
{
  int rc;
  MPI t1;
  MPI t2;
  MPI base[4];
  MPI exp[4];

  if (!(mpi_cmp_ui(a, 0) > 0 && mpi_cmp(a, pkey->p) < 0))
    return 0;                    /* assertion  0 < a < p  failed */

  t1 = mpi_alloc(mpi_get_nlimbs(a));
  t2 = mpi_alloc(mpi_get_nlimbs(a));

#if 0
  /* t1 = (y^a mod p) * (a^b mod p) mod p */
  mpi_powm(t1, pkey->y, a, pkey->p);
  mpi_powm(t2, a, b, pkey->p);
  mpi_mulm(t1, t1, t2, pkey->p);

  /* t2 = g ^ input mod p */
  mpi_powm(t2, pkey->g, input, pkey->p);

  rc = !mpi_cmp(t1, t2);
#elif 0
  /* t1 = (y^a mod p) * (a^b mod p) mod p */
  base[0] = pkey->y;
  exp[0] = a;
  base[1] = a;
  exp[1] = b;
  base[2] = NULL;
  exp[2] = NULL;
  mpi_mulpowm(t1, base, exp, pkey->p);

  /* t2 = g ^ input mod p */
  mpi_powm(t2, pkey->g, input, pkey->p);

  rc = !mpi_cmp(t1, t2);
#else
  /* t1 = g ^ - input * y ^ a * a ^ b  mod p */
  mpi_invm(t2, pkey->g, pkey->p);
  base[0] = t2;
  exp[0] = input;
  base[1] = pkey->y;
  exp[1] = a;
  base[2] = a;
  exp[2] = b;
  base[3] = NULL;
  exp[3] = NULL;
  mpi_mulpowm(t1, base, exp, pkey->p);
  rc = !mpi_cmp_ui(t1, 1);

#endif

  mpi_free(t1);
  mpi_free(t2);
  return rc;
}



/* --------- *
 * Interface *
 * --------- */



CIPHER_ELGAMAL_SKEY *elgamal_gen_secret(unsigned nbits)
{
  CIPHER_ELGAMAL_SKEY *skey;
  
  skey = malloc(sizeof(*skey));
  generate(skey, nbits, 0);  /* We trust the factors */
  return(skey);
}


CIPHER_ELGAMAL_PKEY *elgamal_gen_public(CIPHER_ELGAMAL_SKEY *skey)
{
  CIPHER_ELGAMAL_PKEY *pkey;

  pkey = malloc(sizeof(*pkey));
  pkey->p = mpi_copy(skey->p);
  pkey->g = mpi_copy(skey->g);
  pkey->y = mpi_copy(skey->y);
  return(pkey);
}


CIPHER_ELGAMAL_SKEY *elgamal_dup_secret(CIPHER_ELGAMAL_SKEY *old)
{
  CIPHER_ELGAMAL_SKEY *new;
  
  new = malloc(sizeof(*new));
  new->p = mpi_copy(old->p);
  new->g = mpi_copy(old->g);
  new->y = mpi_copy(old->y);
  new->x = mpi_copy(old->x);
  
  return(new);
}


CIPHER_ELGAMAL_PKEY *elgamal_dup_public(CIPHER_ELGAMAL_PKEY *old)
{
  CIPHER_ELGAMAL_PKEY *new;
  
  new = malloc(sizeof(*new));
  new->p = mpi_copy(old->p);
  new->g = mpi_copy(old->g);
  new->y = mpi_copy(old->y);
  
  return(new);
}


int elgamal_test_secret(CIPHER_ELGAMAL_SKEY *skey)
{
  return(check_secret_key(skey));  /* TRUE or FALSE */
}


int elgamal_bits_public(CIPHER_ELGAMAL_PKEY *pkey)
{
  return(mpi_get_nbits(pkey->p));
}


int elgamal_bits_secret(CIPHER_ELGAMAL_SKEY *skey)
{
  return(mpi_get_nbits(skey->p));
}


int elgamal_is_pair(CIPHER_ELGAMAL_PKEY *pkey, CIPHER_ELGAMAL_SKEY *skey)
{
  if (mpi_cmp(pkey->p, skey->p)) return(0);
  if (mpi_cmp(pkey->g, skey->g)) return(0);
  if (mpi_cmp(pkey->y, skey->y)) return(0);
  return(1);  /* TRUE, they're a pair */
}


void elgamal_sign(CIPHER_ELGAMAL_SKEY *skey,
                  void **signature, size_t *signature_len,
                  void *hash, size_t hash_len)
{
  MPI resarr[2];
  MPI data;
  byte *signature_buf, *temp_buf;
  unsigned int temp_len;
  unsigned int key_len;

  key_len = (elgamal_bits_secret(skey) + 7) >> 3;

  resarr[0] = mpi_alloc(mpi_get_nlimbs(skey->p));
  resarr[1] = mpi_alloc(mpi_get_nlimbs(skey->p));
  data = mpi_alloc_bytes(hash_len);

  /* Encode hashtext -> MPI */

  mpi_set_buffer(data, (const byte *) hash, hash_len, 0);

  /* Sign it */

  sign(resarr[0], resarr[1], data, skey);

  /* Encode MPIs -> signaturetext */

  signature_buf = malloc(key_len << 1);
  memset(signature_buf, 0, key_len << 1);

  temp_buf = mpi_get_buffer(resarr[0], &temp_len, 0);
  memcpy(signature_buf + (key_len - temp_len), temp_buf, temp_len);
  free(temp_buf);

  temp_buf = mpi_get_buffer(resarr[1], &temp_len, 0);
  memcpy(signature_buf + key_len + (key_len - temp_len), temp_buf, temp_len);
  free(temp_buf);

  *signature_len = key_len << 1;
  *signature = signature_buf;

  /* Free scratch variables */
  
  mpi_free(resarr[0]);
  mpi_free(resarr[1]);
  mpi_free(data);
  
  /* *signature = [ab] */
}


int elgamal_verify(CIPHER_ELGAMAL_PKEY *pkey,
                   void *hash, size_t hash_len,
                   void *signature, size_t signature_len)
{
  int ret;
  MPI hash_mpi;
  MPI cip_a, cip_b;

  hash_mpi = mpi_alloc_secure(mpi_get_nlimbs(pkey->p));
  cip_a = mpi_alloc(mpi_get_nlimbs(pkey->p));
  cip_b = mpi_alloc(mpi_get_nlimbs(pkey->p));

  /* Encode signaturetext / hash -> MPIs */

  mpi_set_buffer(cip_a, signature, signature_len >> 1, 0);
  mpi_set_buffer(cip_b, signature + (signature_len >> 1), signature_len >> 1, 0);
  mpi_set_buffer(hash_mpi, hash, hash_len, 0);

  /* Verify it */

  ret = verify(cip_a, cip_b, hash_mpi, pkey);

  /* Free scratch variables */

  mpi_free(cip_a);
  mpi_free(cip_b);
  mpi_free(hash_mpi);

  return(ret);
}


/* TODO: Save execution time by making more friendly mpicoder functions */

void elgamal_encrypt(CIPHER_ELGAMAL_PKEY *pkey,
                     void **cipher, size_t *cipher_len,
                     void *plain, size_t plain_len)
{
  MPI resarr[2];
  MPI data;
  byte *cipher_buf, *temp_buf;
  unsigned int temp_len;
  unsigned int key_len;
  
  key_len = (elgamal_bits_public(pkey) + 7) >> 3;

  resarr[0] = mpi_alloc(mpi_get_nlimbs(pkey->p));
  resarr[1] = mpi_alloc(mpi_get_nlimbs(pkey->p));
  data = mpi_alloc(1);
  
  /* Encode plaintext -> MPI */

  mpi_set_buffer(data, (const byte *) plain, plain_len, 0);
  mpi_set_highbit(data, plain_len << 3);  /* Preserve leading zeros */

  /* Encrypt it */

  encrypt(resarr[0], resarr[1], data, pkey);

  /* Encode MPIs -> ciphertext */

  cipher_buf = malloc(key_len << 1);
  memset(cipher_buf, 0, key_len << 1);

  temp_buf = mpi_get_buffer(resarr[0], &temp_len, 0);
  memcpy(cipher_buf + (key_len - temp_len), temp_buf, temp_len);
  free(temp_buf);

  temp_buf = mpi_get_buffer(resarr[1], &temp_len, 0);
  memcpy(cipher_buf + key_len + (key_len - temp_len), temp_buf, temp_len);
  free(temp_buf);

  *cipher_len = key_len << 1;
  *cipher = cipher_buf;

  /* Free scratch variables */
  
  mpi_free(resarr[0]);
  mpi_free(resarr[1]);
  mpi_free(data);
  
  /* *cipher = [ab] */
}


void elgamal_decrypt(CIPHER_ELGAMAL_SKEY *skey,
                     void **plain, size_t *plain_len,
                     void *cipher, size_t cipher_len)
{
  MPI result;
  MPI cip_a, cip_b;
  int nbytes;

  result = mpi_alloc_secure(mpi_get_nlimbs(skey->p));
  cip_a = mpi_alloc(mpi_get_nlimbs(skey->p));
  cip_b = mpi_alloc(mpi_get_nlimbs(skey->p));

  /* Encode ciphertext -> MPIs */

  mpi_set_buffer(cip_a, cipher, (cipher_len + 1) >> 1, 0);
  mpi_set_buffer(cip_b, cipher + ((cipher_len + 1) >> 1), (cipher_len + 1) >> 1, 0);

  /* Decrypt it */

  decrypt(result, cip_a, cip_b, skey);

  /* Encode MPI -> plaintext */

  mpi_normalize(result);  /* Remove leading non-data zeros */
  nbytes = (mpi_get_nbits(result) + 6) >> 3;  /* .. - 1 ? */

  *plain_len = nbytes;
  mpi_clear_highbit(result, mpi_get_nbits(result) - 1);  /* Remove protection */
  *plain = mpi_get_buffer_with_size(result, plain_len, 0);

  /* Free scratch variables */

  mpi_free(cip_a);
  mpi_free(cip_b);
  mpi_free(result);
}


void elgamal_export_secret(CIPHER_ELGAMAL_SKEY *skey, byte **out, unsigned int *len)
{
  unsigned int p_len, g_len, y_len, x_len;
  unsigned l;
  byte *buf;

  p_len = (mpi_get_nbits(skey->p) + 7) >> 3;
  g_len = (mpi_get_nbits(skey->g) + 7) >> 3;
  y_len = (mpi_get_nbits(skey->y) + 7) >> 3;
  x_len = (mpi_get_nbits(skey->x) + 7) >> 3;

  *len = 8 + p_len + g_len + y_len + x_len;

  *out = malloc(*len);
  **out = p_len >> 8;
  *(*out + 1) = p_len & 0xff;
  *(*out + 2) = g_len >> 8;
  *(*out + 3) = g_len & 0xff;
  *(*out + 4) = y_len >> 8;
  *(*out + 5) = y_len & 0xff;
  *(*out + 6) = x_len >> 8;
  *(*out + 7) = x_len & 0xff;

  buf = mpi_get_buffer(skey->p, &l, 0);
  memcpy(*out + 8, buf, p_len);
  free(buf);

  buf = mpi_get_buffer(skey->g, &l, 0);
  memcpy(*out + 8 + p_len, buf, g_len);
  free(buf);

  buf = mpi_get_buffer(skey->y, &l, 0);
  memcpy(*out + 8 + p_len + g_len, buf, y_len);
  free(buf);

  buf = mpi_get_buffer(skey->x, &l, 0);
  memcpy(*out + 8 + p_len + g_len + y_len, buf, x_len);
  free(buf);
}


CIPHER_ELGAMAL_SKEY *elgamal_import_secret(const byte *in)
{
  CIPHER_ELGAMAL_SKEY *skey;
  unsigned int p_len, g_len, y_len, x_len;

  p_len = (in[0] << 8) + in[1];
  g_len = (in[2] << 8) + in[3];
  y_len = (in[4] << 8) + in[5];
  x_len = (in[6] << 8) + in[7];
  
  skey = malloc(sizeof(*skey));
  skey->p = mpi_alloc_bytes(p_len);
  skey->g = mpi_alloc_bytes(g_len);
  skey->y = mpi_alloc_bytes(y_len);
  skey->x = mpi_alloc_bytes(x_len);

  mpi_set_buffer(skey->p, in + 8, p_len, 0);
  mpi_set_buffer(skey->g, in + 8 + p_len, g_len, 0);
  mpi_set_buffer(skey->y, in + 8 + p_len + g_len, y_len, 0);
  mpi_set_buffer(skey->x, in + 8 + p_len + g_len + y_len, x_len, 0);
  
  return(skey);
}


void elgamal_export_public(CIPHER_ELGAMAL_PKEY *pkey, byte **out, unsigned int *len)
{
  unsigned int p_len, g_len, y_len;
  unsigned l;
  byte *buf;

  p_len = (mpi_get_nbits(pkey->p) + 7) >> 3;
  g_len = (mpi_get_nbits(pkey->g) + 7) >> 3;
  y_len = (mpi_get_nbits(pkey->y) + 7) >> 3;

  *len = 8 + p_len + g_len + y_len;

  *out = malloc(*len);
  **out = p_len >> 8;
  *(*out + 1) = p_len & 0xff;
  *(*out + 2) = g_len >> 8;
  *(*out + 3) = g_len & 0xff;
  *(*out + 4) = y_len >> 8;
  *(*out + 5) = y_len & 0xff;

  buf = mpi_get_buffer(pkey->p, &l, 0);
  memcpy(*out + 8, buf, p_len);
  free(buf);

  buf = mpi_get_buffer(pkey->g, &l, 0);
  memcpy(*out + 8 + p_len, buf, g_len);
  free(buf);

  buf = mpi_get_buffer(pkey->y, &l, 0);
  memcpy(*out + 8 + p_len + g_len, buf, y_len);
  free(buf);
}


CIPHER_ELGAMAL_PKEY *elgamal_import_public(const byte *in)
{
  CIPHER_ELGAMAL_PKEY *pkey;
  unsigned int p_len, g_len, y_len;

  p_len = (in[0] << 8) + in[1];
  g_len = (in[2] << 8) + in[3];
  y_len = (in[4] << 8) + in[5];
  
  pkey = malloc(sizeof(*pkey));
  pkey->p = mpi_alloc_bytes(p_len);
  pkey->g = mpi_alloc_bytes(g_len);
  pkey->y = mpi_alloc_bytes(y_len);

  mpi_set_buffer(pkey->p, in + 8, p_len, 0);
  mpi_set_buffer(pkey->g, in + 8 + p_len, g_len, 0);
  mpi_set_buffer(pkey->y, in + 8 + p_len + g_len, y_len, 0);

  return(pkey);
}


void elgamal_del_secret(CIPHER_ELGAMAL_SKEY *skey)
{
  memset(skey, 0, sizeof(*skey));
  free(skey);
}


void elgamal_del_public(CIPHER_ELGAMAL_PKEY *pkey)
{
  free(pkey);
}
