/* Copyright (C) 1999-2000 Hans Petter K. Jansson
 *
 * This library is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
 *
 * You can contact the library's author by sending e-mail to <hpj@styx.net>.
 */

#include "config.h"
#include "types.h"
#include "log.h"
#include "mpi.h"
#include "random.h"
#include "cipher.h"

#include <stdlib.h>


/****************
 * Return true if n is probably a prime
 */
static int is_prime(MPI n, int steps, int *count)
{
  MPI x = mpi_alloc( mpi_get_nlimbs(n));
  MPI y = mpi_alloc( mpi_get_nlimbs(n));
  MPI z = mpi_alloc( mpi_get_nlimbs(n));
  MPI nminus1 = mpi_alloc(mpi_get_nlimbs(n));
  MPI a2 = mpi_alloc_set_ui(2);
  MPI q;
  unsigned i, j, k;
  int rc = 0;
  unsigned nbits = mpi_get_nbits(n);


  mpi_sub_ui(nminus1, n, 1);
  
  /* find q and k, so that n = 1 + 2^k * q */
  q = mpi_copy(nminus1);
  k = mpi_trailing_zeros(q);
  mpi_tdiv_q_2exp(q, q, k);

  for (i = 0; i < steps; i++)
  {
    ++*count;

    if (!i)
    {
      mpi_set_ui(x, 2);
    }
    else
    {
      /*mpi_set_bytes( x, nbits-1, get_random_byte, 0 );*/
      {
        char *p = rand_bits_dyn(nbits);
        mpi_set_buffer(x, p, (nbits + 7) / 8, 0);
        free(p);
      }
      
      /* make sure that the number is smaller than the prime
       * and keep the randomness of the high bit */
      if (mpi_test_bit(x, nbits - 2))
      {
        mpi_set_highbit(x, nbits - 2);  /* clear all higher bits */
      }
      else
      {
        mpi_set_highbit(x, nbits - 2);
        mpi_clear_bit(x, nbits - 2);
      }
/*
      assert(mpi_cmp(x, nminus1) < 0 && mpi_cmp_ui(x, 1) > 0);
 */
    }
    
    mpi_powm(y, x, q, n);

    if (mpi_cmp_ui(y, 1) && mpi_cmp(y, nminus1))
    {
      for (j = 1; j < k && mpi_cmp(y, nminus1); j++)
      {
        mpi_powm(y, y, a2, n);
        if (!mpi_cmp_ui(y, 1)) goto leave;  /* not a prime */
      }
      if (mpi_cmp(y, nminus1)) goto leave;  /* not a prime */
    }
  }
  rc = 1; /* may be a prime */

  leave:

  mpi_free(x);
  mpi_free(y);
  mpi_free(z);
  mpi_free(nminus1);
  mpi_free(q);
  mpi_free(a2);

  return rc;
}


MPI cipher_prime_gen(unsigned nbits)
{
  unsigned nlimbs;
  MPI prime, val_2, val_3, result;
  int i;
  unsigned x, step;
  unsigned count1, count2;
  int *mods;


  /* Count small primes in table. */

  if (!_prime_small_num)
  {
    for (i=0; _prime_small[i]; i++) _prime_small_num++;
  }


  mods = malloc(_prime_small_num * sizeof *mods);

  /* make nbits fit into MPI implementation */
  nlimbs = (nbits + BITS_PER_MPI_LIMB - 1) /  BITS_PER_MPI_LIMB;
  val_2  = mpi_alloc(nlimbs);
  mpi_set_ui(val_2, 2);
  val_3  = mpi_alloc(nlimbs);
  mpi_set_ui(val_3, 3);
  result = mpi_alloc(nlimbs);
  prime  = /* secret ? mpi_alloc_secure(nlimbs) : */ mpi_alloc(nlimbs);
  count1 = count2 = 0;

  /* Enter (endless) loop. It is only broken by an eventual return. */

  for (;;)
  {
    /* generate a random number */

    /* mpi_set_bytes(prime, nbits, get_random_byte, randomlevel); */
    {
      char *p = rand_bits_dyn(nbits);
      mpi_set_buffer(prime, p, (nbits + 7) / 8, 0);
      free(p);
    }

    /* set high order bit to 1, set low order bit to 1 */
    mpi_set_highbit(prime, nbits - 1);
    mpi_set_bit(prime, 0);

    /* calculate all remainders */
    for(i = 0; (x = _prime_small[i]); i++)
      mods[i] = mpi_fdiv_r_ui(NULL, prime, x);

    for (step = 0; step < 20000; step += 2)
    {
      /* check against all the small primes we have in mods */
      count1++;

      for(i=0; (x = _prime_small[i]); i++)
      {
        while (mods[i] + step >= x) mods[i] -= x;
        if (!(mods[i] + step)) break;
      }

      if (x) continue;  /* found a multiple of a already known prime */

      mpi_add_ui(prime, prime, step);

      /* perform stronger tests */

      if (is_prime(prime, 5, &count2))
      {
        if (!mpi_test_bit(prime, nbits - 1))
        {
        }

        mpi_free(val_2);
        mpi_free(val_3);
        mpi_free(result);
        free(mods);
        return prime;
      }
    }
  }
}


/****************
 * Returns: true if this may be a prime
 */
int cipher_prime_check(MPI prime, MPI val_2)
{
  int i;
  unsigned x;
  int count = 0;

  /* check against small primes */
  for(i = 0; (x = _prime_small[i]); i++)
  {
    if (mpi_divisible_ui(prime, x)) return 0;
  }

  /* a quick fermat test */
  {
    MPI result = mpi_alloc_like(prime);
    MPI pminus1 = mpi_alloc_like(prime);

    mpi_sub_ui(pminus1, prime, 1);
    mpi_powm(result, val_2, pminus1, prime);
    mpi_free(pminus1);
    
    if (mpi_cmp_ui(result, 1))
    {
      /* if composite */
      mpi_free(result);
      return 0;
    }
    mpi_free(result);
  }

  /* perform stronger tests */
  if (is_prime(prime, 5, &count)) return 1; /* is probably a prime */
  return 0;
}


void m_out_of_n(char *array, int m, int n)
{
  int i = 0, i1 = 0, j = 0, jp = 0,  j1 = 0, k1 = 0, k2 = 0;

  if (!m || m >= n) return;

  if (m == 1)
  {
    /* special case */
    for (i = 0; i < n; i++)
      if (array[i])
      {
        array[i++] = 0;
        if (i >= n) i = 0;
        array[i] = 1;
        return;
      }
    log_put_abort("Cipher bug #2.");
  }

  for (j = 1; j < n; j++)
  {
    if (array[n - 1] == array[n - j - 1]) continue;
    j1 = j;
    break;
  }

  if (m & 1)
  {
    /* m is odd */
    if (array[n-1])
    {
      if (j1 & 1)
      {
        k1 = n - j1;
        k2 = k1 + 2;
        if (k2 > n) k2 = n;
        goto leave;
      }
      goto scan;
    }
    
    k2 = n - j1 - 1;
    if (k2 == 0)
    {
      k1 = i;
      k2 = n - j1;
    }
    else if (array[k2] && array[k2 - 1]) k1 = n;
    else k1 = k2 + 1;
  }
  else
  {
    /* m is even */
    if (!array[n - 1])
    {
      k1 = n - j1;
      k2 = k1 + 1;
      goto leave;
    }

    if (!(j1 & 1))
    {
      k1 = n - j1;
      k2 = k1+2;
      if (k2 > n) k2 = n;
      goto leave;
    }
    
    scan:
    
    jp = n - j1 - 1;
    for (i=1; i <= jp; i++)
    {
      i1 = jp + 2 - i;
      if (array[i1 - 1])
      {
        if (array[i1 - 2])
        {
          k1 = i1 - 1;
          k2 = n - j1;
        }
        else
        {
          k1 = i1 - 1;
          k2 = n + 1 - j1;
        }
        goto leave;
      }
    }
    k1 = 1;
    k2 = n + 1 - m;
  }
  
  leave:
  
  array[k1 - 1] = !array[k1 - 1];
  array[k2 - 1] = !array[k2 - 1];
}
