/* Copyright (C) 2000 Andrei Zmievski
 *
 * This library is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
 *
 * You can contact the library's author by sending e-mail to
 * <andrei@ispi.net>.
 */

#include "config.h"
#include "types.h"
#include "bmem.h"
#include <stdlib.h>
#include <string.h>

#ifdef __GNUC__
__inline
#endif
static BCHUNK *bmem_chunk_new(BMEM *mem)
{
  BCHUNK *chunk;

  chunk = (BCHUNK*)malloc(sizeof(BCHUNK) + mem->storage_size);
  memset((void *) chunk, (int) 0, (size_t) sizeof(BCHUNK));
  memset(chunk->bitmap, 0xff, BALLOC_MEM_CHUNKS/8);
  chunk->start_looking_from_index = 0;

  return(chunk);
}

#ifdef __GNUC__
__inline
#endif
static int bmem_chunk_del(void *key, void *data)
{
  free((BCHUNK *)data);

  return 0;
}

BMEM *bmem_new(size_t block_size)
{
  BMEM *mem;

  mem = (BMEM *)malloc(sizeof(BMEM));
  memset(mem, 0, sizeof(BMEM));

  mem->eligible_chunks = bt_new_without_bmem();
  mem->saturated_chunks = bt_new_without_bmem();
  mem->block_size = block_size;
  mem->storage_size = block_size * BALLOC_MEM_CHUNKS;
  mem->current_chunk = bmem_chunk_new(mem);
  bt_insert(mem->eligible_chunks, mem->current_chunk, mem->current_chunk);
  mem->num_eligible_chunks = 1;

  return(mem);
}

void bmem_del(BMEM *mem)
{
  bt_for_all(mem->eligible_chunks, bmem_chunk_del);
  bt_for_all(mem->saturated_chunks, bmem_chunk_del);
  bt_del(mem->eligible_chunks);
  bt_del(mem->saturated_chunks);
  free(mem);
}


#define BLOCK_OFFSET (mem->block_size * ((bitmap_index << BALLOC_MULSHIFT_AMOUNT) + word_bit))

void *bmem_block_new(BMEM *mem)
{
  BCHUNK *chunk;
  ulong bitmap_word;
  u16 word_bit;
  u16 bitmap_index;
  ulong *bitmap;

  chunk = mem->current_chunk;

  /* --- Find free index --- */

  bitmap_index = chunk->start_looking_from_index;
  word_bit = chunk->start_looking_from_bit;

  bitmap = chunk->bitmap;
  bitmap_word = bitmap[bitmap_index];
  if (!bitmap_word)
  {
    word_bit = 0;

    for (bitmap_index++; ; bitmap_index++)
    {
      if (!bitmap[bitmap_index]) continue;
      bitmap_word = bitmap[bitmap_index];
      break;
    }
  }

  for (bitmap_word >>= word_bit; !(bitmap_word & 1); bitmap_word >>= 1, word_bit++) ;

  bitmap[bitmap_index] ^= (ulong) 1 << word_bit;

  /* This sets up the start_index for the next time */
  if (bitmap[bitmap_index] >> 1)
  {
    chunk->start_looking_from_index = bitmap_index;
    chunk->start_looking_from_bit = word_bit + 1;
  }
  else
  {
    if (bitmap_index == BALLOC_BITMAP_SIZE - 1)
      chunk->start_looking_from_index = 0;
    else
      chunk->start_looking_from_index = bitmap_index + 1;
    
    chunk->start_looking_from_bit = 0;
  }

  /* --- Check allocation status and return result --- */
  
  /* If chunk is saturated, get a new one. */
  if (++chunk->allocated == BALLOC_MEM_CHUNKS)
  {
    bt_move(mem->saturated_chunks, mem->eligible_chunks, chunk);
    mem->num_eligible_chunks--;

    if (!mem->num_eligible_chunks)
    {
      mem->current_chunk = bmem_chunk_new(mem);
      bt_insert(mem->eligible_chunks, (void *) mem->current_chunk, (void *) mem->current_chunk);
      mem->num_eligible_chunks++;
    }
    else
      mem->current_chunk = bt_get_root(mem->eligible_chunks);
  }

  return((byte *) BALLOC_CHUNK_STORAGE(chunk) + BLOCK_OFFSET);
}


#define max_bound block

void bmem_block_del(BMEM *mem, void *block)
{
  BCHUNK *chunk;
  u16 freed_block_index;
  u16 chunk_was_saturated = 0;
  u16 bitmap_index, word_bit;
  void *min_bound;
  
  min_bound = (void *) ((psize) block - sizeof(BCHUNK) - mem->storage_size);
  

  chunk = (BCHUNK *) bt_find_first_in_range(mem->eligible_chunks, (void *) min_bound, max_bound);

  if (!chunk) 
  {
    chunk = (BCHUNK *) bt_find_first_in_range(mem->saturated_chunks, (void *) min_bound, max_bound);
    chunk_was_saturated = 1;
  }

  freed_block_index = (psize) ((psize) block - (psize) BALLOC_CHUNK_STORAGE(chunk)) / (psize) mem->block_size;

#if SIZEOF_UNSIGNED_LONG == 4
  bitmap_index = (freed_block_index >> 5);  /* / 32 */
  word_bit = (freed_block_index & 0x1f);    /* [0 .. 31] */
#elif SIZEOF_UNSIGNED_LONG == 8
  bitmap_index = (freed_block_index >> 6);  /* / 64 */
  word_bit = (freed_block_index & 0x3f);    /* [0 .. 63] */
#else
#error Unsupported architecture.
#endif

  if (bitmap_index < chunk->start_looking_from_index)
  {
    chunk->start_looking_from_index = bitmap_index;
    chunk->start_looking_from_bit = word_bit;
  }
  else if (word_bit < chunk->start_looking_from_bit)
    chunk->start_looking_from_bit = word_bit;

  chunk->bitmap[bitmap_index] ^= (ulong) 1 << word_bit;
  chunk->allocated--;

  if (!chunk->allocated && mem->num_eligible_chunks > 1)
  {
    bt_remove(mem->eligible_chunks, chunk);
    mem->num_eligible_chunks--;
    free(chunk);
  }
  else if (chunk_was_saturated)
  {
    bt_move(mem->eligible_chunks, mem->saturated_chunks, chunk);
    mem->num_eligible_chunks++;
  }
}
