/* Copyright (C) 1999-2000 Hans Petter K. Jansson
 *
 * This library is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
 *
 * You can contact the library's author by sending e-mail to <hpj@styx.net>.
 */

#include "config.h"
#include "tt.h"
#include "sock.h"
#include "nvtp.h"

#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/mman.h>


void nvtp_pull(NVTP *nvtp)
{
  unsigned char *p;
  int len, dequeued_len;
  int i;
  unsigned char c;

  _sock_feed(nvtp->s);

  /* This is what we're expecting next */

  while (fifobuf_used(nvtp->s->fib_in))
  {
    switch (nvtp->state_in)
    {
      case NVTP_STATE_LINE_DATA:
        if (!nvtp->root_in) nvtp->root_in = tt_new();

        dequeued_len = 0;

        FIFOBUF_FOR_DATA(nvtp->s->fib_in, 4096, p, len)
        {
          /* p   is pointer to chunk
           * len is length of chunk */

          for (i = 0; i < len; i++)
          {
            if (*(p + i) == '\n')
            {
              /* Copy data up to LF */
              tt_data_append_bytes(nvtp->root_in, p, i);
              tt_add_as_last_child(nvtp->queue_in, nvtp->root_in);
              nvtp->root_in = 0;
              nvtp->state_in = NVTP_STATE_MAYBE_CR;
              break;
            }
            else if (*(p + i) == '\r')
            {
              /* Copy data up to CR */
              tt_data_append_bytes(nvtp->root_in, p, i);
              tt_add_as_last_child(nvtp->queue_in, nvtp->root_in);
              nvtp->root_in = 0;
              nvtp->state_in = NVTP_STATE_MAYBE_LF;
              break;
            }
          }

          dequeued_len += i;

          if (i == len) tt_data_append_bytes(nvtp->root_in, p, len);
          else break;
        }

        if (nvtp->state_in != NVTP_STATE_LINE_DATA)
          fifobuf_drop(nvtp->s->fib_in, dequeued_len + 1);
        else if (dequeued_len)
          fifobuf_drop(nvtp->s->fib_in, dequeued_len);

        break;

      case NVTP_STATE_MAYBE_CR:
        fifobuf_peek(nvtp->s->fib_in, &c, 1);
        if (c == '\r') fifobuf_drop(nvtp->s->fib_in, 1);
        nvtp->state_in = NVTP_STATE_LINE_DATA;
        break;

      case NVTP_STATE_MAYBE_LF:
        fifobuf_peek(nvtp->s->fib_in, &c, 1);
        if (c == '\n') fifobuf_drop(nvtp->s->fib_in, 1);
        nvtp->state_in = NVTP_STATE_LINE_DATA;
        break;

      case NVTP_STATE_NONE:
      default:
        /* This could mean the comm is uninitialized; deal with it later */
        nvtp->state_in = NVTP_STATE_LINE_DATA;
        break;
    }
  }
}


void nvtp_push(NVTP *nvtp)
{
  unsigned int size;
  byte *p;
  char cr = '\r';
  char lf = '\n';

  _sock_spill(nvtp->s);
  
  /* This is what we're expecting next */

  while (fifobuf_free_max(nvtp->s->fib_out) &&
         (tt_count_children(nvtp->queue_out) || nvtp->root_out ||
          nvtp->state_out != NVTP_STATE_LINE_DATA))  /* FIXME: Slightly ugly */
  {
    switch (nvtp->state_out)
    {
      case NVTP_STATE_LINE_DATA:
        if (!nvtp->root_out)
        {
          nvtp->root_out = tt_get_first_child(nvtp->queue_out);
          tt_detach(tt_get_first_child(nvtp->queue_out));
          nvtp->byte_out = 0;
          /* nvtp->root_out must be set according to loop criteria */
        }
      
        size = fifobuf_free_max(nvtp->s->fib_out) <
               (tt_size(nvtp->root_out) - nvtp->byte_out) ?
               fifobuf_free_max(nvtp->s->fib_out) :
               (tt_size(nvtp->root_out) - nvtp->byte_out);

        p = tt_data_get(nvtp->root_out) + nvtp->byte_out;  /* FIXME: Not file compatible */
        fifobuf_enqueue(nvtp->s->fib_out, p, size);
        nvtp->byte_out += size;
      
        if (nvtp->byte_out == tt_size(nvtp->root_out))
        {
          tt_del(nvtp->root_out);
          nvtp->root_out = 0;
          nvtp->byte_out = 0;
          nvtp->state_out = NVTP_STATE_CR;
        }

        break;

      case NVTP_STATE_CR:
        fifobuf_enqueue(nvtp->s->fib_out, &cr, 1);
        nvtp->state_out = NVTP_STATE_LF;
        break;
      
      case NVTP_STATE_LF:
        fifobuf_enqueue(nvtp->s->fib_out, &lf, 1);
        nvtp->state_out = NVTP_STATE_LINE_DATA;
        break;

      case NVTP_STATE_NONE:
      default:
        nvtp->state_out = NVTP_STATE_LINE_DATA;
        break;
    }
  }
  
  return;
}
