/* Copyright (C) 1999-2000 Hans Petter K. Jansson
 *
 * This library is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
 *
 * You can contact the library's author by sending e-mail to <hpj@styx.net>.
 */

#include "config.h"
#include "proxy.h"
#include "tt.h"
#include "nvtp.h"

#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/time.h>
#ifdef TIME_WITH_SYS_TIME
# include <time.h>
#endif


/* ----------------- *
 * Check NVTP queues *
 * ----------------- */


void proxy_dequeue_nvtps(PROXY *p)
{
  TT *tt0, *tt1;
  NVTP *nvtp;

  /* --- Check nvtp queues --- */
  
  TT_FOR_EACH(p->nvtps, tt0)
  {
    nvtp = (NVTP *) tt_data_get_ptr(tt0);
    while (nvtp_queue_in_len(nvtp))
    {
      tt1 = nvtp_dequeue_tt(nvtp);
      proxy_call_nvtp(p, nvtp, tt1);
      tt_del(tt1);
    }
  }
}


/* -------------------------------- *
 * Set up proxy file descriptor map *
 * -------------------------------- */


void proxy_setup_nvtps(PROXY *p)
{
  TT *tt0;
  int fd;

  /* --- Find all nvtps --- */
  
  TT_FOR_EACH(p->nvtps, tt0)
  {
    if (nvtp_is_active(tt_data_get_ptr(tt0)))
    {
      fd = nvtp_get_fd_read(tt_data_get_ptr(tt0));
      if (fd > p->fd_max) p->fd_max = fd;
      FD_SET(fd, &(p->fd_read));
      
      if (nvtp_buf_out_size(tt_data_get_ptr(tt0)) ||
          nvtp_is_connecting(tt_data_get_ptr(tt0)))
      {
        fd = nvtp_get_fd_write(tt_data_get_ptr(tt0));
        FD_SET(fd, &(p->fd_write));
        if (fd > p->fd_max) p->fd_max = fd;
      }
    }
  }
}


/* ------------------------------- *
 * Check NVTPs for recent activity *
 * ------------------------------- */


void proxy_check_nvtps(PROXY *p, fd_set *fd_read, fd_set *fd_write)
{
  TT *tt0, *tt1;
  NVTP *nvtp, *nvtp_child;

  /* --- Check nvtps --- */
  
  TT_FOR_EACH(p->nvtps, tt0)
  {
    nvtp = (NVTP *) tt_data_get_ptr(tt0);
    
    /* Write */
    
    if (FD_ISSET(nvtp_get_fd_write(nvtp), fd_write))
    {
      if (nvtp_is_connected(nvtp)) nvtp_push(nvtp);
      else  /* if (comm_is_connecting(c)) */
      {
        /* Connection attempt resolved */
        nvtp_child = nvtp_accept(nvtp);
        
        /* FIXME: if (c_child && !connect_handler) comm_del(c_child) */
        if (nvtp_child && (tt1 = tt_find_first_child(tt0, "connect", 7)))
          proxy_call_nvtp_connect(nvtp_child, tt1);
      }
    }
    
    /* Read */
    
    if (FD_ISSET(nvtp_get_fd_read(nvtp), fd_read))
    {
      if (nvtp_is_connected(nvtp))
      {
        if (_sock_feed(nvtp_get_sock(nvtp)) < 1)
        {
          /* Disconnected (EOF) */
          if ((tt1 = tt_find_first_child(tt0, "close", 5)))
            proxy_call_nvtp_close(nvtp, tt1);
          else
          {
            proxy_del_nvtp(p, nvtp);
            nvtp_del(nvtp);
          }
        }
        else
        {
          nvtp_pull(nvtp);
          while (nvtp_queue_in_len(nvtp))
          {
            tt1 = nvtp_dequeue_tt(nvtp);
            proxy_call_nvtp(p, nvtp, tt1);
            if (p->changed && !tt_is_parent_of(p->nvtps, tt0)) break;
          }
        }
      }
      else /* if (comm_is_accepting(s)) */
      {
        /* Incoming connection (probably) */
        nvtp_child = nvtp_accept(nvtp);
        
        /* FIXME: if (c_child && !connect_handler) comm_del(c_child) */
        if (nvtp_child && (tt1 = tt_find_first_child(tt0, "connect", 7)))
          proxy_call_nvtp_connect(nvtp_child, tt1);
      }
    }
    
    if (p->changed) break;
  }
}


/* ------ *
 * Adding *
 * ------ */

/* <comm> -> <completeness> -> root   -> <root>
 *                          -> trans  -> <trans>
 *                          -> recv   -> <func>  -> <priv>
 *        -> <completeness> -> ...
 *        -> connect        -> <func> -> <priv>
 *        -> close          -> <func> -> <priv>
 *        -> ...                                 */


void proxy_add_nvtp_connect(PROXY *p, NVTP *nvtp, void *priv,
                            int (*connect)(NVTP *, void *))
{
  TT *tt0, *tt1;
  
  p->changed = 1;

  tt0 = tt_find_first_child(p->nvtps, &nvtp, sizeof(nvtp));
  if (!tt0) tt0 = tt_new_with_parent_and_data(p->nvtps, &nvtp, sizeof(nvtp));

  tt1 = tt_find_first_child(tt0, "connect", 7);
  if (tt1) tt_del(tt1);
  tt1 = tt_new_with_parent_and_data(tt0, "connect", 7);

  tt1 = tt_new_with_parent_and_data(tt1, &connect, sizeof(connect));
  tt_new_with_parent_and_data(tt1, &priv, sizeof(priv));
}


void proxy_add_nvtp_close(PROXY *p, NVTP *nvtp, void *priv,
                          int (*close)(NVTP *, void *))
{
  TT *tt0, *tt1;
  
  p->changed = 1;

  tt0 = tt_find_first_child(p->nvtps, &nvtp, sizeof(nvtp));
  if (!tt0) tt0 = tt_new_with_parent_and_data(p->nvtps, &nvtp, sizeof(nvtp));

  tt1 = tt_find_first_child(tt0, "close", 5);
  if (tt1) tt_del(tt1);
  tt1 = tt_new_with_parent_and_data(tt0, "close", 5);

  tt1 = tt_new_with_parent_and_data(tt1, &close, sizeof(close));
  tt_new_with_parent_and_data(tt1, &priv, sizeof(priv));
}


void proxy_add_nvtp_in(PROXY *p, NVTP *nvtp, void *priv,
                       int (*in)(NVTP *, void *))
{
  TT *tt0, *tt1;
  
  p->changed = 1;

  tt0 = tt_find_first_child(p->nvtps, &nvtp, sizeof(nvtp));
  if (!tt0) tt0 = tt_new_with_parent_and_data(p->nvtps, &nvtp, sizeof(nvtp));

  tt1 = tt_find_first_child(tt0, "in", 2);
  if (tt1) tt_del(tt1);
  tt1 = tt_new_with_parent_and_data(tt0, "in", 2);

  tt1 = tt_new_with_parent_and_data(tt1, &in, sizeof(in));
  tt_new_with_parent_and_data(tt1, &priv, sizeof(priv));
}


/* -------- *
 * Removing *
 * -------- */


void proxy_del_nvtp(PROXY *p, NVTP *nvtp)
{
  TT *tt0;

  if ((tt0 = tt_find_first_child(p->nvtps, &nvtp, sizeof(nvtp))))
  {
    /* Should only need doing once.
       NOTE: Forgot what the above sentence refers to. */

    p->changed = 1;
    tt_del(tt0);
  }
}


/* ------- *
 * Calling *
 * ------- */


void proxy_call_nvtp(PROXY *p, NVTP *nvtp, TT *tt)
{
  TT *tt0, *tt1;
  int (*pcn_recv)(NVTP *, TT *, void *);
  void *priv;

  tt0 = tt_find_first_child(p->nvtps, &nvtp, sizeof(nvtp));
  if (!tt0) return;

  tt1 = tt_find_first_child(tt0, "in", 2);
  if (!tt1) return;
  
  tt1 = tt_get_first_child(tt1);
  if (!tt1) return;

  pcn_recv = tt_data_get_ptr(tt1);
  
  tt1 = tt_get_first_child(tt1);
  if (tt1) priv = tt_data_get_ptr(tt1);
  else priv = 0;

  pcn_recv(nvtp, tt, priv);
  tt_del(tt);
}


void proxy_call_nvtp_connect(NVTP *nvtp, TT *tt)
{
  int (*func_nvtp)(NVTP *, void *);
  void *priv;

  tt = tt_get_first_child(tt);
  func_nvtp = tt_data_get_ptr(tt);
  tt = tt_get_first_child(tt);
  if (tt) priv = tt_data_get_ptr(tt);
  else priv = 0;

  func_nvtp(nvtp, priv);
}


void proxy_call_nvtp_close(NVTP *nvtp, TT *tt)
{
  int (*func_nvtp)(NVTP *, void *);
  void *priv;

  tt = tt_get_first_child(tt);
  func_nvtp = tt_data_get_ptr(tt);
  tt = tt_get_first_child(tt);
  if (tt) priv = tt_data_get_ptr(tt);
  else priv = 0;

  func_nvtp(nvtp, priv);
}
