/* Copyright (C) 1999-2000 Hans Petter K. Jansson
 *
 * This library is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
 *
 * You can contact the library's author by sending e-mail to <hpj@styx.net>.
 */

#include "config.h"
#include "ftime.h"
#include "tt.h"
#include "private.h"

#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/time.h>
#ifdef TIME_WITH_SYS_TIME
# include <time.h>
#endif

#include <errno.h>


/* ---------------- *
 * Proxy allocation *
 * ---------------- */


PROXY *proxy_new()
{
  PROXY *p;
  
  p = malloc(sizeof(*p));
  memset(p, 0, sizeof(*p));

  p->socks = tt_new_with_data("sock", 4);
  p->nvtps = tt_new_with_data("nvtp", 4);
  p->smtps = tt_new_with_data("smtp", 4);
  p->comms = tt_new_with_data("comm", 4);
  p->timers = tt_new_with_data("time", 4);
  p->signals = tt_new_with_data("sig", 3);
  p->changed = 1;
  return(p);
}


void proxy_del(PROXY *p)
{
  tt_del(p->socks);
  tt_del(p->nvtps);
  tt_del(p->smtps);
  tt_del(p->comms);
  tt_del(p->timers);
  tt_del(p->signals);
  free(p);
}


/* ------------- *
 * Proxy running *
 * ------------- */


void proxy_signal_catch(int sig)
{
  /* TODO: Make this function. This is not as trivial as it might seem,
   *       as we'll need a global (threadsafe) registry of proxies so
   *       that all of them can catch a signal when it arrives. */

  /*  
    TT_FOR_EACH(
  */
}


void proxy_loop(PROXY *p)
{
  TT *tt0, *tt1;
  int fd_num;
  struct timeval tv0, *tv1, tv2, tv4;
  fd_set fd_read, fd_write;

  p->exiting = 0;

  while (!p->exiting)
  {
    proxy_dequeue_nvtps(p);
    proxy_dequeue_smtps(p);
    proxy_dequeue_comms(p);

    /* --- Update filedes list if needed --- */
    
    p->changed = 1;  /* TODO: Update changed flag when buffers empty */

    if (p->changed)
    {
      FD_ZERO(&(p->fd_read));
      FD_ZERO(&(p->fd_write));
      p->fd_max = -1;

      proxy_setup_socks(p);
      proxy_setup_nvtps(p);
      proxy_setup_smtps(p);
      proxy_setup_comms(p);

      p->fd_max += 1;
      p->changed = 0;
    }

    /* --- Find closest timeout --- */

    gettimeofday(&tv0, 0);
    tv4.tv_usec = 0;
    tv4.tv_sec = 0x0000ffff;

    TT_FOR_EACH(p->timers, tt0)
    {
      tt1 = tt_get_first_child(tt_find_first_child(tt0, "last", 4));
      tv1 = (struct timeval *) tt_data_get(tt1);
      memcpy(&tv2, tv1, sizeof(tv2));

      tt1 = tt_get_first_child(tt_find_first_child(tt0, "interval", 8));
      tv1 = (struct timeval *) tt_data_get(tt1);
      timeradd(&tv2, tv1, &tv2);

      /* Now: tv2 is absolute time, at which to trigger */

      timersub(&tv2, &tv0, &tv2);

      /* Now: tv2 is time from now, at which to trigger */

      if (timercmp(&tv2, &tv4, <)) memcpy(&tv4, &tv2, sizeof(tv4));

      /* Now: tv4 is closest timeout found */
    }

    /* --- Wait for something to happen --- */

    memcpy(&fd_read, &(p->fd_read), sizeof(fd_read));
    memcpy(&fd_write, &(p->fd_write), sizeof(fd_write));
    fd_num = select(p->fd_max, &fd_read, &fd_write, 0, &tv4);
    
    /* --- Check comms, nvtps and sockets --- */

    if (fd_num > 0)
    {
      proxy_check_nvtps(p, &fd_read, &fd_write);
      proxy_check_smtps(p, &fd_read, &fd_write);
      proxy_check_comms(p, &fd_read, &fd_write);
      proxy_check_socks(p, &fd_read, &fd_write);
    }

    /* --- Check timeouts --- */

    gettimeofday(&tv0, 0);
    tv4.tv_sec = 0x7ffffffe;

    TT_FOR_EACH(p->timers, tt0)
    {
      tt1 = tt_get_first_child(tt_find_first_child(tt0, "last", 4));
      tv1 = (struct timeval *) tt_data_get(tt1);
      memcpy(&tv2, tv1, sizeof(tv2));

      tt1 = tt_get_first_child(tt_find_first_child(tt0, "interval", 8));
      tv1 = (struct timeval *) tt_data_get(tt1);
      timeradd(&tv2, tv1, &tv2);

      /* Now: tv2 is absolute time, at which to trigger */
      
      if (timercmp(&tv2, &tv0, <)) proxy_call_timer(p, tt0);
    }
  }
}


void proxy_break(PROXY *p)
{
  p->exiting = 1;
}
