/* Copyright (C) 1999-2000 Hans Petter K. Jansson
 *
 * This library is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
 *
 * You can contact the library's author by sending e-mail to <hpj@styx.net>.
 */

#include "config.h"
#include "proxy.h"
#include "tt.h"
#include "smtp.h"

#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/time.h>
#ifdef TIME_WITH_SYS_TIME
# include <time.h>
#endif


/* ----------------- *
 * Check SMTP queues *
 * ----------------- */


void proxy_dequeue_smtps(PROXY *p)
{
  TT *tt0, *tt1;
  SMTP *smtp;

  /* --- Check smtp queues --- */
  
  TT_FOR_EACH(p->smtps, tt0)
  {
    smtp = (SMTP *) tt_data_get_ptr(tt0);
    while (smtp_queue_in_len(smtp))
    {
      tt1 = smtp_dequeue(smtp);
      proxy_call_smtp(p, smtp, tt1);
      tt_del(tt1);
    }
  }
}


/* -------------------------------- *
 * Set up proxy file descriptor map *
 * -------------------------------- */


void proxy_setup_smtps(PROXY *p)
{
  TT *tt0;
  int fd;

  /* --- Find all smtps --- */
  
  TT_FOR_EACH(p->smtps, tt0)
  {
    if (smtp_is_active(tt_data_get_ptr(tt0)))
    {
      fd = smtp_get_fd_read(tt_data_get_ptr(tt0));
      if (fd > p->fd_max) p->fd_max = fd;
      FD_SET(fd, &(p->fd_read));
      
      if (nvtp_queue_out_len(smtp_get_nvtp(tt_data_get_ptr(tt0))) ||
          smtp_buf_out_size(tt_data_get_ptr(tt0)) ||
          smtp_is_connecting(tt_data_get_ptr(tt0)))
      {
        fd = smtp_get_fd_write(tt_data_get_ptr(tt0));
        FD_SET(fd, &(p->fd_write));
        if (fd > p->fd_max) p->fd_max = fd;
      }
    }
  }
}


/* ------------------------------- *
 * Check SMTPs for recent activity *
 * ------------------------------- */


void proxy_check_smtps(PROXY *p, fd_set *fd_read, fd_set *fd_write)
{
  TT *tt0, *tt1;
  SMTP *smtp, *smtp_child;

  /* --- Check smtps --- */
  
  TT_FOR_EACH(p->smtps, tt0)
  {
    smtp = (SMTP *) tt_data_get_ptr(tt0);
    
    /* Write */
    
    if (FD_ISSET(smtp_get_fd_write(smtp), fd_write))
    {
      if (smtp_is_connected(smtp)) smtp_push(smtp);
      else  /* if (comm_is_connecting(c)) */
      {
        /* Connection attempt resolved */
        smtp_child = smtp_accept(smtp);
        
        /* FIXME: if (c_child && !connect_handler) comm_del(c_child) */
        if (smtp_child && (tt1 = tt_find_first_child(tt0, "connect", 7)))
          proxy_call_smtp_connect(smtp_child, tt1);
      }
    }
    
    /* Read */
    
    if (FD_ISSET(smtp_get_fd_read(smtp), fd_read))
    {
      if (smtp_is_connected(smtp))
      {
        if (_sock_feed(smtp_get_sock(smtp)) < 1)
        {
          /* Disconnected (EOF) */
          if ((tt1 = tt_find_first_child(tt0, "close", 5)))
            proxy_call_smtp_close(smtp, tt1);
          else
          {
            proxy_del_smtp(p, smtp);
            smtp_del(smtp);
          }
        }
        else
        {
          smtp_pull(smtp);
          while (smtp_queue_in_len(smtp))
          {
            tt1 = smtp_dequeue(smtp);
            proxy_call_smtp(p, smtp, tt1);
            if (p->changed && !tt_is_parent_of(p->smtps, tt0)) break;
          }
        }
      }
      else /* if (comm_is_accepting(s)) */
      {
        /* Incoming connection (probably) */
        smtp_child = smtp_accept(smtp);
        
        /* FIXME: if (c_child && !connect_handler) comm_del(c_child) */
        if (smtp_child && (tt1 = tt_find_first_child(tt0, "connect", 7)))
          proxy_call_smtp_connect(smtp_child, tt1);
      }
    }
    
    if (p->changed) break;
  }
}


/* ------ *
 * Adding *
 * ------ */

/* <comm> -> <completeness> -> root   -> <root>
 *                          -> trans  -> <trans>
 *                          -> recv   -> <func>  -> <priv>
 *        -> <completeness> -> ...
 *        -> connect        -> <func> -> <priv>
 *        -> close          -> <func> -> <priv>
 *        -> ...                                 */


void proxy_add_smtp_connect(PROXY *p, SMTP *smtp, void *priv,
                            int (*connect)(SMTP *, void *))
{
  TT *tt0, *tt1;
  
  p->changed = 1;

  tt0 = tt_find_first_child(p->smtps, &smtp, sizeof(smtp));
  if (!tt0) tt0 = tt_new_with_parent_and_data(p->smtps, &smtp, sizeof(smtp));

  tt1 = tt_find_first_child(tt0, "connect", 7);
  if (tt1) tt_del(tt1);
  tt1 = tt_new_with_parent_and_data(tt0, "connect", 7);

  tt1 = tt_new_with_parent_and_data(tt1, &connect, sizeof(connect));
  tt_new_with_parent_and_data(tt1, &priv, sizeof(priv));
}


void proxy_add_smtp_close(PROXY *p, SMTP *smtp, void *priv,
                          int (*close)(SMTP *, void *))
{
  TT *tt0, *tt1;
  
  p->changed = 1;

  tt0 = tt_find_first_child(p->smtps, &smtp, sizeof(smtp));
  if (!tt0) tt0 = tt_new_with_parent_and_data(p->smtps, &smtp, sizeof(smtp));

  tt1 = tt_find_first_child(tt0, "close", 5);
  if (tt1) tt_del(tt1);
  tt1 = tt_new_with_parent_and_data(tt0, "close", 5);

  tt1 = tt_new_with_parent_and_data(tt1, &close, sizeof(close));
  tt_new_with_parent_and_data(tt1, &priv, sizeof(priv));
}


void proxy_add_smtp_in(PROXY *p, SMTP *smtp, void *priv,
                       int (*in)(SMTP *, TT *, void *))
{
  TT *tt0, *tt1;
  
  p->changed = 1;

  tt0 = tt_find_first_child(p->smtps, &smtp, sizeof(smtp));
  if (!tt0) tt0 = tt_new_with_parent_and_data(p->smtps, &smtp, sizeof(smtp));

  tt1 = tt_find_first_child(tt0, "in", 2);
  if (tt1) tt_del(tt1);
  tt1 = tt_new_with_parent_and_data(tt0, "in", 2);

  tt1 = tt_new_with_parent_and_data(tt1, &in, sizeof(in));
  tt_new_with_parent_and_data(tt1, &priv, sizeof(priv));
}


/* -------- *
 * Removing *
 * -------- */


void proxy_del_smtp(PROXY *p, SMTP *smtp)
{
  TT *tt0;

  if ((tt0 = tt_find_first_child(p->smtps, &smtp, sizeof(smtp))))
  {
    /* Should only need doing once.
       NOTE: Forgot what the above sentence refers to. */

    p->changed = 1;
    tt_del(tt0);
  }
}


/* ------- *
 * Calling *
 * ------- */


void proxy_call_smtp(PROXY *p, SMTP *smtp, TT *tt)
{
  TT *tt0, *tt1;
  int (*pcn_recv)(SMTP *, TT *, void *);
  void *priv;

  tt0 = tt_find_first_child(p->smtps, &smtp, sizeof(smtp));
  if (!tt0) return;

  tt1 = tt_find_first_child(tt0, "in", 2);
  if (!tt1) return;
  
  tt1 = tt_get_first_child(tt1);
  if (!tt1) return;

  pcn_recv = tt_data_get_ptr(tt1);
  
  tt1 = tt_get_first_child(tt1);
  if (tt1) priv = tt_data_get_ptr(tt1);
  else priv = 0;

  pcn_recv(smtp, tt, priv);
  tt_del(tt);
}


void proxy_call_smtp_connect(SMTP *smtp, TT *tt)
{
  int (*func_smtp)(SMTP *, void *);
  void *priv;

  tt = tt_get_first_child(tt);
  func_smtp = tt_data_get_ptr(tt);
  tt = tt_get_first_child(tt);
  if (tt) priv = tt_data_get_ptr(tt);
  else priv = 0;

  func_smtp(smtp, priv);
}


void proxy_call_smtp_close(SMTP *smtp, TT *tt)
{
  int (*func_smtp)(SMTP *, void *);
  void *priv;

  tt = tt_get_first_child(tt);
  func_smtp = tt_data_get_ptr(tt);
  tt = tt_get_first_child(tt);
  if (tt) priv = tt_data_get_ptr(tt);
  else priv = 0;

  func_smtp(smtp, priv);
}
