/* Copyright (C) 1999-2000 Hans Petter K. Jansson
 *
 * This library is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
 *
 * You can contact the library's author by sending e-mail to <hpj@styx.net>.
 */

#include "config.h"
#include "proxy.h"
#include "sock.h"

#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/time.h>
#ifdef TIME_WITH_SYS_TIME
# include <time.h>
#endif


/* -------------------------------- *
 * Set up proxy file descriptor map *
 * -------------------------------- */


void proxy_setup_socks(PROXY *p)
{
  TT *tt0;
  int fd;

  /* --- Find all sockets --- */

  TT_FOR_EACH(p->socks, tt0)
  {
    if (sock_is_active(tt_data_get_ptr(tt0)))
    {
      fd = sock_get_fd_read(tt_data_get_ptr(tt0));
      FD_SET(fd, &(p->fd_read));
      if (fd > p->fd_max) p->fd_max = fd;
      
      if (sock_buf_out_size(tt_data_get_ptr(tt0)) ||
          sock_is_connecting(tt_data_get_ptr(tt0)) ||
          tt_find_first_child(tt0, "allwrites", 9))
      {
        fd = sock_get_fd_write(tt_data_get_ptr(tt0));
        FD_SET(fd, &(p->fd_write));
        if (fd > p->fd_max) p->fd_max = fd;
      }
    }
  }
}


/* --------------------------------- *
 * Check sockets for recent activity *
 * --------------------------------- */


void proxy_check_socks(PROXY *p, fd_set *fd_read, fd_set *fd_write)
{
  TT *tt0, *tt1;
  SOCK *s, *s_child;

  /* --- Check sockets --- */

  TT_FOR_EACH(p->socks, tt0)
  {
    s = (SOCK *) tt_data_get_ptr(tt0);
    
    /* Write */
    
    if (FD_ISSET(sock_get_fd_write(s), fd_write))
    {
      if (sock_is_connected(s))
      {
        /* Outgoing data (hungry socket) */
        _sock_spill(s);
        if ((tt1 = tt_find_first_child(tt0, "write", 5)))
          proxy_call_sock(s, tt1);
      }
      else /* if (sock_is_connecting(s)) */
      {
        /* Connection attempt resolved */
        s_child = sock_accept(s);
        
        if (s_child && (tt1 = tt_find_first_child(tt0, "connect", 7)))
          proxy_call_sock_connect(s_child, tt1);
      }
    }
    
    if (p->changed) break;
    
    /* Read */
    
    if (FD_ISSET(sock_get_fd_read(s), fd_read))
    {
      if (sock_is_connected(s))
      {
        if (_sock_feed(s) < 1)
        {
          /* Disconnected (EOF) */
          if ((tt1 = tt_find_first_child(tt0, "close", 5)))
            proxy_call_sock(s, tt1);
          else
            if (!sock_disconnect(s, 0)) proxy_del_sock(p, s);
        }
        else if ((tt1 = tt_find_first_child(tt0, "read", 4)))
        {
          /* Incoming data */
          proxy_call_sock(s, tt1);
        }
      }
      else /* if (sock_is_accepting(s)) */
      {
        /* Incoming connection (probably) */
        s_child = sock_accept(s);
        
        if (s_child && (tt1 = tt_find_first_child(tt0, "connect", 7)))
          proxy_call_sock_connect(s_child, tt1);
      }
    }
    
    if (p->changed) break;
  }
}


/* ------ *
 * Adding *
 * ------ */


void proxy_add_sock_read(PROXY *p, SOCK *s, void *priv,
                         int (*read)(SOCK *, void *))
{
  TT *tt0, *tt1;
  
  tt0 = tt_find_first_child(p->socks, &s, sizeof(s));
  if (!tt0) tt0 = tt_new_with_parent_and_data(p->socks, &s, sizeof(s));

  tt1 = tt_find_first_child(tt0, "read", 4);
  if (tt1) tt_del(tt1);
  tt1 = tt_new_with_parent_and_data(tt0, "read", 4);
  tt1 = tt_new_with_parent_and_data(tt1, &read, sizeof(read));
  tt_new_with_parent_and_data(tt1, &priv, sizeof(priv));
}


void proxy_add_sock_write(PROXY *p, SOCK *s, void *priv,
                          int (*write)(SOCK *, void *))
{
  TT *tt0, *tt1;
  
  tt0 = tt_find_first_child(p->socks, &s, sizeof(s));
  if (!tt0) tt0 = tt_new_with_parent_and_data(p->socks, &s, sizeof(s));

  tt1 = tt_find_first_child(tt0, "write", 5);
  if (tt1) tt_del(tt1);
  tt1 = tt_new_with_parent_and_data(tt0, "write", 5);
  tt1 = tt_new_with_parent_and_data(tt1, &write, sizeof(write));
  tt_new_with_parent_and_data(tt1, &priv, sizeof(priv));
}


void proxy_add_sock_connect(PROXY *p, SOCK *s, void *priv,
                            int (*connect)(SOCK *, void *))
{
  TT *tt0, *tt1;
  
  tt0 = tt_find_first_child(p->socks, &s, sizeof(s));
  if (!tt0) tt0 = tt_new_with_parent_and_data(p->socks, &s, sizeof(s));

  tt1 = tt_find_first_child(tt0, "connect", 7);
  if (tt1) tt_del(tt1);
  tt1 = tt_new_with_parent_and_data(tt0, "connect", 7);
  tt1 = tt_new_with_parent_and_data(tt1, &connect, sizeof(connect));
  tt_new_with_parent_and_data(tt1, &priv, sizeof(priv));
}


void proxy_add_sock_close(PROXY *p, SOCK *s, void *priv,
                          int (*close)(SOCK *, void *))
{
  TT *tt0, *tt1;
  
  tt0 = tt_find_first_child(p->socks, &s, sizeof(s));
  if (!tt0) tt0 = tt_new_with_parent_and_data(p->socks, &s, sizeof(s));

  tt1 = tt_find_first_child(tt0, "close", 5);
  if (tt1) tt_del(tt1);
  tt1 = tt_new_with_parent_and_data(tt0, "close", 5);
  tt1 = tt_new_with_parent_and_data(tt1, &close, sizeof(close));
  tt_new_with_parent_and_data(tt1, &priv, sizeof(priv));
}


void proxy_add_sock(PROXY *p, SOCK *s, void *priv,
                    int (*read)(SOCK *, void *),
                    int (*write)(SOCK *, void *),
                    int (*connect)(SOCK *, void *),
                    int (*close)(SOCK *, void *))
{
  TT *tt0, *tt1;

  p->changed = 1;

  tt0 = tt_find_first_child(p->socks, &s, sizeof(s));
  if (!tt0) tt0 = tt_new_with_parent_and_data(p->socks, &s, sizeof(s));

  if (read)
  {
    tt1 = tt_find_first_child(tt0, "read", 4);
    if (tt1) tt_del(tt1);
    tt1 = tt_new_with_parent_and_data(tt0, "read", 4);
    tt1 = tt_new_with_parent_and_data(tt1, &read, sizeof(read));
    tt_new_with_parent_and_data(tt1, &priv, sizeof(priv));
  }

  if (write)
  {
    tt1 = tt_find_first_child(tt0, "write", 5);
    if (tt1) tt_del(tt1);
    tt1 = tt_new_with_parent_and_data(tt0, "write", 5);
    tt1 = tt_new_with_parent_and_data(tt1, &write, sizeof(write));
    tt_new_with_parent_and_data(tt1, &priv, sizeof(priv));
  }

  if (connect)
  {
    tt1 = tt_find_first_child(tt0, "connect", 7);
    if (tt1) tt_del(tt1);
    tt1 = tt_new_with_parent_and_data(tt0, "connect", 7);
    tt1 = tt_new_with_parent_and_data(tt1, &connect, sizeof(connect));
    tt_new_with_parent_and_data(tt1, &priv, sizeof(priv));
  }

  if (close)
  {
    tt1 = tt_find_first_child(tt0, "close", 5);
    if (tt1) tt_del(tt1);
    tt1 = tt_new_with_parent_and_data(tt0, "close", 5);
    tt1 = tt_new_with_parent_and_data(tt1, &close, sizeof(close));
    tt_new_with_parent_and_data(tt1, &priv, sizeof(priv));
  }
}


/* -------- *
 * Removing *
 * -------- */


void proxy_del_sock(PROXY *p, SOCK *s)
{
  TT *tt0;
  
  tt0 = tt_find_first_child(p->socks, &s, sizeof(s));
  if (tt0)
  {
    p->changed = 1;
    tt_del(tt0);
  }
}


/* ------- *
 * Calling *
 * ------- */


void proxy_call_sock(SOCK *s, TT *tt)
{
  int (*func_sock)(SOCK *, void *);
  void *priv;

  tt = tt_get_first_child(tt);
  func_sock = tt_data_get_ptr(tt);
  tt = tt_get_first_child(tt);
  if (tt) priv = tt_data_get_ptr(tt);
  else priv = 0;

  func_sock(s, priv);
}


void proxy_call_sock_connect(SOCK *s, TT *tt)
{
  int (*func_sock)(SOCK *, void *);
  void *priv;

  tt = tt_get_first_child(tt);
  func_sock = tt_data_get_ptr(tt);
  tt = tt_get_first_child(tt);
  if (tt) priv = tt_data_get_ptr(tt);
  else priv = 0;

  func_sock(s, priv);
}


/* -------- *
 * Blocking *
 * -------- */


void proxy_unblock_sock_writes(PROXY *p, SOCK *s)
{
  TT *tt0;

  if (!(tt0 = tt_find_first_child(p->socks, &s, sizeof(s)))) return;
  tt_new_with_parent_and_data(tt0, "allwrites", 9);
  p->changed = 1;
}


void proxy_block_sock_writes(PROXY *p, SOCK *s)
{
  TT *tt0;

  if (!(tt0 = tt_find_first_child(p->socks, &s, sizeof(s)))) return;
  if (!(tt0 = tt_find_first_child(tt0, "allwrites", 9))) return;
  tt_del(tt0);
  p->changed=1;
}
