/* Copyright (C) 1999-2000 Hans Petter K. Jansson
 *
 * This library is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
 *
 * You can contact the library's author by sending e-mail to <hpj@styx.net>.
 */

#include "config.h"
#include "fstring.h"
#include "tt.h"
#include "smtp.h"
#include "nvtp.h"
#include <stdlib.h>


/* --- Generating commands --- */


void smtp_cmd_hello(SMTP *smtp, char *host)
{
  TT *tt;
  
  tt = tt_new_with_data("HELO ", 5);
  if (host) tt_data_append_bytes(tt, host, strlen(host));
  else tt_data_append_bytes(tt, "anonymous.org", 13);
  nvtp_enqueue_tt(smtp_get_nvtp(smtp), tt);
}


void smtp_cmd_mail(SMTP *smtp, char *from)
{
  TT *tt;

  tt = tt_new_with_data("MAIL FROM:<", 11);
  tt_data_append_bytes(tt, from, strlen(from));
  tt_data_append_bytes(tt, ">", 1);
  nvtp_enqueue_tt(smtp_get_nvtp(smtp), tt);
}


void smtp_cmd_recipient(SMTP *smtp, char *recipient)
{
  TT *tt;

  tt = tt_new_with_data("RCPT TO:<", 9);
  tt_data_append_bytes(tt, recipient, strlen(recipient));
  tt_data_append_bytes(tt, ">", 1);
  nvtp_enqueue_tt(smtp_get_nvtp(smtp), tt);
}


void smtp_cmd_data(SMTP *smtp)
{
  nvtp_enqueue_tt(smtp_get_nvtp(smtp), tt_new_with_data("DATA", 4));
}


void smtp_cmd_noop(SMTP *smtp)
{
  nvtp_enqueue_tt(smtp_get_nvtp(smtp), tt_new_with_data("NOOP", 4));
}


void smtp_cmd_reset(SMTP *smtp)
{
  nvtp_enqueue_tt(smtp_get_nvtp(smtp), tt_new_with_data("RSET", 4));
}


void smtp_cmd_quit(SMTP *smtp)
{
  nvtp_enqueue_tt(smtp_get_nvtp(smtp), tt_new_with_data("QUIT", 4));
}


/* --- Identifying commands --- */


TT *smtp_parse_arg_address(TT *cmd)
{
  int c, c1;
  char buf[8], *dbuf;
  TT *arg;

  c = tt_spn_from(cmd, " ", 4);       /* Skip spaces after command */
  if (c < 0) return(0);               /* Arg error */
  c = tt_cspn_from(cmd, ":", c);      /* Look for ':' in "FROM:" / "TO:" */
  if (c < 0) return(0);               /* Arg error */
  c = tt_spn_from(cmd, " ", c + 1);   /* Skip spaces after "FROM:" / "TO:" */
  if (c < 0) return(0);               /* Arg error */
  tt_data_get_bytes(cmd, buf, c, 1);
  if (buf[0] == '<')
  {
    c++;
    c1 = tt_cspn_from(cmd, ">", c);
    if (c1 < 0) return(0);
  }
  else
    c1 = tt_size(cmd);
  
  /* c  indexes first character in address
     c1 indexes last character in address + 1 */

  dbuf = malloc(c1 - c);
  tt_data_get_bytes(cmd, dbuf, c, c1 - c);
  arg = tt_new_with_data(dbuf, c1 - c);  /* Could use internal with copy=FALSE here */
  free(dbuf);
  
  return(arg);
}


int smtp_parse_cmd(TT *tt, TT **arg)
{
  char buf[8];

  *arg = 0;
  if (!tt_has_data(tt)) return(SMTP_CMD_BLANK);
  if (tt_size(tt) < 4) return(SMTP_CMD_UNKNOWN);

  tt_data_get_bytes(tt, buf, 0, 4);
  memlwr(buf, 4);

  /* Commands without an argument */
  if (!memcmp(buf, "helo", 4)) return(SMTP_CMD_HELLO);  /* Has arg, but don't care */
  if (!memcmp(buf, "noop", 4)) return(SMTP_CMD_NOOP);
  if (!memcmp(buf, "rset", 4)) return(SMTP_CMD_RESET);
  if (!memcmp(buf, "quit", 4)) return(SMTP_CMD_QUIT);
  if (!memcmp(buf, "data", 4)) return(SMTP_CMD_DATA);

  /* Commands with one argument */
  if (!memcmp(buf, "mail", 4))
  {
    *arg = smtp_parse_arg_address(tt);
    return(SMTP_CMD_MAIL);
  }
  
  if (!memcmp(buf, "rcpt", 4))
  {
    *arg = smtp_parse_arg_address(tt);
    return(SMTP_CMD_RECIPIENT);
  }
  
  return(SMTP_CMD_UNKNOWN);
}
