/* Copyright (C) 1999-2000 Hans Petter K. Jansson
 *
 * This library is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
 *
 * You can contact the library's author by sending e-mail to <hpj@styx.net>.
 */

#include "config.h"
#include "tt.h"
#include "smtp.h"
#include "nvtp.h"

#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/mman.h>


/* Mail TT structure - an example:
 * 
 * mail -> sender     -> daniel@b.cl
 *
 *      -> recipients -> hpj@styx.net
 *                    -> joakim@styx.net
 *                    -> ...
 * 
 *      -> headers    -> subject -> "Tomb of the Unknown Implementors"
 * 
 *                    -> cc      -> joakim@styx.net -> "Joakim Ziegler"
 *                               -> torfinn@nome.no -> "Torfinn Nome"
 *                               -> ...
 *
 *      -> body -> "Experience is hard-earned.\r\nAlways.\r\nDid you know..."
 */


void smtp_pull(SMTP *smtp)
{
  TT *tt, *arg;
  int code;

  nvtp_pull(smtp_get_nvtp(smtp));

  while (nvtp_queue_in_len(smtp_get_nvtp(smtp)))
  {
    if (smtp->role == SMTP_ROLE_CLIENT)
    {
      tt = nvtp_dequeue_tt(smtp->nvtp);
      code = smtp_parse_reply(smtp, tt);
      
      if (code < 400)
      {
        /* Success */
        smtp->state_in = SMTP_STATE_IDLE;
      }
      else
      {
        /* Failure; find out what failed */

        switch (smtp->state_in)
        {
          case SMTP_STATE_HELLO:
            /* Our HELO failed. Big deal - see what happens */
            break;
          
          case SMTP_STATE_MAIL:
            /* Couldn't send mail */
            break;
          
          case SMTP_STATE_RECIPIENT:
            /* Recipient not accepted */
            break;
          
          case SMTP_STATE_DATA_BEGIN:
            /* DATA statement was rejected */
            break;
          
          case SMTP_STATE_DATA_END:
            /* BODY was rejected (probably internal server error) */
            break;

          case SMTP_STATE_NOOP:
            /* Server was unable to do nothing */
            break;
          
          case SMTP_STATE_RESET:
            /* Reset failed (shouldn't happen) */
            break;

          case SMTP_STATE_QUIT:
            /* Quit failed (let's quit anyway) */
            break;

          case SMTP_STATE_IDLE:
            /* Got a reply without sending command; TURN?
             * Anyhow, we don't care. */
            switch (code)
            {
              case SMTP_REPLY_SERVICE_UNAVAILABLE:
                /* Server doesn't like us */
                /* FIXME: Order this smtp to close */
                break;
              case SMTP_REPLY_CLOSING_CHANNEL:
                /* FIXME: Order this smtp to close */
                break;
              default:
                /* Unhandled error code */
                break;
            }

            break;

          default:
            smtp->state_in = SMTP_STATE_IDLE;
            break;
        }
      }
      
      tt_del(tt);
    }
    else  /* smtp->role == SMTP_ROLE_SERVER */
    {
      tt = nvtp_dequeue_tt(smtp->nvtp);

      switch (smtp->state_in)
      {
        case SMTP_STATE_COMMAND_INIT:
          /* helo, mail, noop, rset, quit */
          switch (smtp_parse_cmd(tt, &arg))
          {
            /* Allowed */
            
            case SMTP_CMD_HELLO:
              smtp_reply(smtp, 250, "anonymous.org");  /* FIXME: Get/set host name */
              break;
            
            case SMTP_CMD_MAIL:
              if (arg)
              {
                smtp_reply(smtp, 250, "Ok");
                smtp->root_in = tt_new_with_data("mail", 4);
                smtp->node_in = tt_new_with_parent_and_data(smtp->root_in, "sender", 6);
                tt_add(smtp->node_in, arg);
                smtp->node_in = tt_new_with_parent_and_data(smtp->root_in, "recipients", 10);
                smtp->state_in = SMTP_STATE_COMMAND_MAIL;
              }
              else
                smtp_reply(smtp, 501, "Syntax error: MAIL FROM: <address>");
              break;

            case SMTP_CMD_NOOP:
              smtp_reply(smtp, 250, "Ok");
              break;

            case SMTP_CMD_RESET:
              smtp_reply(smtp, 250, "Ok");
              break;

            case SMTP_CMD_QUIT:
              smtp_reply(smtp, 221, "Bye.");
              /* FIXME: Queue socket close command */
              break;
              
              /* Disallowed */
              
            case SMTP_CMD_RECIPIENT:
              smtp_reply(smtp, 503, "Error: Need MAIL command.");
              break;
              
            case SMTP_CMD_DATA:
              smtp_reply(smtp, 503, "Error: Need MAIL then RCPT commands.");
              break;
              
            case SMTP_CMD_UNKNOWN:
              smtp_reply(smtp, 502, "Error: Command not implemented.");
            default:
              break;
          }

          tt_del(tt);
          break;
          
        case SMTP_STATE_COMMAND_MAIL:
          /* helo, noop, rcpt, data, rset, quit */
          switch (smtp_parse_cmd(tt, &arg))
          {
            /* Allowed */

            case SMTP_CMD_HELLO:
              smtp_reply(smtp, 250, "anonymous.org");  /* FIXME: Get/set host name */
              break;

            case SMTP_CMD_NOOP:
              smtp_reply(smtp, 250, "Ok");
              break;

            case SMTP_CMD_RECIPIENT:
              if (arg)
              {
                smtp_reply(smtp, 250, "Ok");
                tt_add_as_last_child(smtp->node_in, arg);
              }
              else
                smtp_reply(smtp, 501, "Syntax error: RCPT TO: <address>");
              break;
              
            case SMTP_CMD_DATA:
              /* TODO: Must have recipients */
              if (!tt_is_parent(smtp->node_in))
                smtp_reply(smtp, 503, "Error: Need RCPT commands.");
              else
              {
                smtp_reply(smtp, 354, "End data with <CR><LF>.<CR><LF>");
                smtp->node_in = tt_new_with_parent_and_data(smtp->root_in, "headers", 7);
                smtp->state_in = SMTP_STATE_DATA_HEAD;
              }
              break;

            case SMTP_CMD_RESET:
              smtp_reply(smtp, 250, "Ok");
              tt_del(smtp->root_in);  /* Junk the mail in progress */
              break;

            case SMTP_CMD_QUIT:
              smtp_reply(smtp, 221, "Bye.");
              tt_del(smtp->root_in);  /* Junk the mail in progress */
              /* FIXME: Queue socket close command */
              break;

            /* Disallowed */

            case SMTP_CMD_MAIL:
              smtp_reply(smtp, 503, "Error: Nested MAIL command.");
              if (arg) tt_del(arg);
              break;

            case SMTP_CMD_UNKNOWN:
              smtp_reply(smtp, 502, "Error: Command not implemented.");
            default:
              break;
          }

          tt_del(tt);
          break;

        case SMTP_STATE_DATA_HEAD:
          /* <header>, <blank> (start of DATA_BODY) or "." (end of mail) */
          if (!tt_size(tt))
          {
            /* End of headers */
            tt_del(tt);
            smtp->node_in = tt_new_with_parent_and_data(smtp->root_in, "body", 4);
            tt_add(smtp->node_in, tt_new());
            smtp->node_in = tt_get_first_child(smtp->node_in);
            smtp->state_in = SMTP_STATE_DATA_BODY;
          }
          else if (!tt_memcmp(tt, ".", 1))
          {
            /* End of data; what we have is data, not headers */
            tt_del(tt);
            tt = smtp->node_in;
              
            /* Join up what we thought was headers, and assign as body */
            smtp->node_in = tt_new_with_parent_and_data(smtp->root_in, "body", 4);
              
            while (tt_count_children(tt) > 1)
            {
              tt_data_append_bytes(tt_get_first_child(tt), "\r\n", 2);
              tt_join(tt_get_first_child(tt), tt_get_next(tt_get_first_child(tt)));
            }
             
            tt = tt_get_first_child(tt);
            if (tt)
            {
              tt_detach(tt);
              tt_add(smtp->node_in, tt);
            }

            smtp_reply(smtp, 250, "Ok: Queued.");

            /* Queue it up */
            tt_add_as_last_child(smtp->queue_in, smtp->root_in);
            smtp->root_in = 0;
            smtp->state_in = SMTP_STATE_COMMAND_INIT;
          }
          else
            tt_add_as_last_child(smtp->node_in, tt);
          break;

        case SMTP_STATE_DATA_BODY:
          /* <data> or "." (end of mail) */

          if (!tt_memcmp(tt, ".", 1))
          {
            /* End of data */
            smtp_reply(smtp, 250, "Ok: Queued.");

            tt_del(tt);
            tt_add_as_last_child(smtp->queue_in, smtp->root_in);
            smtp->root_in = 0;
            smtp->state_in = SMTP_STATE_COMMAND_INIT;
          }
          else
          { 
            tt_join(smtp->node_in, tt);
            tt_data_append_bytes(smtp->node_in, "\r\n", 2);
          }
          break;
          
        default:
          smtp->state_in = SMTP_STATE_COMMAND_INIT;
          break;
      } /* switch (smtp->state_in) */
    } /* smtp->role == SMTP_ROLE_SERVER */
  } /* while (nvtp_queue_in_len(smtp_get_nvtp(smtp))) */
}


void smtp_push(SMTP *smtp)
{
  char *str;
  TT *tt;

  nvtp_push(smtp_get_nvtp(smtp));

  while (smtp_queue_out_len(smtp))
  {
    if (smtp->role == SMTP_ROLE_CLIENT)
    {
      if (smtp->state_in != SMTP_STATE_IDLE) return;  /* Lockstep */

      switch (smtp->state_out)
      {
        case SMTP_STATE_IDLE:
          /* Right; Process next root from out queue */
          smtp->root_out = tt_get_first_child(smtp->queue_out);
          str = tt_data_get_str(tt_get_first_child(tt_find_first_child(
                                smtp->root_out, "sender", 6)));
          smtp_cmd_mail(smtp, str);
          free(str);
        
          smtp->node_out = tt_get_first_child(tt_find_first_child(
                             smtp->root_out, "recipients", 10));
          smtp->state_out = SMTP_STATE_RECIPIENT;
          smtp->state_in = SMTP_STATE_MAIL;
          break;
        
        case SMTP_STATE_RECIPIENT:
          /* Put the recipient pointed to by node_out, and either
           * a) update to point to next, or
           * b) if there is no next, go to next state (DATA_BEGIN) */
          if (!smtp->node_out) smtp->state_out = SMTP_STATE_DATA_BEGIN;
          {
            str = tt_data_get_str(smtp->node_out);
            smtp_cmd_recipient(smtp, str);
            free(str);
            smtp->node_out = tt_get_next(smtp->node_out);
            smtp->state_in = SMTP_STATE_RECIPIENT;
          }

          break;
        
        case SMTP_STATE_DATA_BEGIN:
          /* Send DATA statement and start waiting for ack */
          smtp_cmd_data(smtp);
          smtp->state_out = SMTP_STATE_DATA_END;
          smtp->state_in = SMTP_STATE_DATA_BEGIN;
          break;

        case SMTP_STATE_DATA_END:
          /* Send data itself, then ".", and start waiting for ack
           * (go back to idle state_out) */
          tt = tt_find_first_child(smtp->root_out, "headers", 7);
          
          break;
      }
    }
    else  /* smtp->role == SMTP_ROLE_SERVER */
    {
      tt = tt_get_first_child(smtp->queue_out);
      tt_detach(tt);
      nvtp_enqueue_tt(smtp_get_nvtp(smtp), tt);
    }
  }
}
