/* Copyright (C) 1999-2000 Hans Petter K. Jansson
 *
 * This library is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
 *
 * You can contact the library's author by sending e-mail to <hpj@styx.net>.
 */

#include "config.h"
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>
#include <fcntl.h>
#include <sys/time.h>
#include <sys/ioctl.h>
#include <sys/socket.h>
#include <sys/file.h>
#include <sys/types.h>
#include <fcntl.h>
#include "log.h"
#include "fifobuf.h"
#include "sock.h"
#include "fstring.h"
#include "cipher.h"


#ifndef MAX
# define MAX(a, b) ((a) > (b) ? (a) : (b))
#endif

#define TRIES 1


/* Necessary prototoypes */

void sock_cipher_pull(SOCK *s);


/* ------------- *
 * Low-level I/O *
 * ------------- */

#ifdef __GNUC__
__inline
#endif
int _sock_feed(SOCK *s)
{
  int t;
  int fd;
  int i0, i1;
  register FIFOBUF *fib;

  fd = SOCK_FDREAD(s);
  fib = s->fib_in;
  i0 = fifobuf_free_cur(fib);
  i1 = fib->bufsize - fib->byte_in;

  if (!i1)
  {
    fib->byte_in = 0;
    fib->node_in = fib->node_in->next;
    i1 = fib->bufsize;
  }

  fcntl(fd, F_SETFL, O_NONBLOCK);
  t = read(fd, fib->node_in->data + fib->byte_in, i0 < i1 ? i0 : i1);

  if (t < 0)
  {
    if (errno != EWOULDBLOCK) _sock_chkerrno(s);
  }
  else
  {
    fib->byte_in += t;
    fib->enqueued += t;

    /* If in cipher negotiation, steal bytes */
    
    if (s->cip && s->cip_state != SOCK_CIP_STATE_DONE) sock_cipher_pull(s);
  }

  fcntl(fd, F_SETFL, 0);
  return(t);  /* Used to detect disconnects */
}


#ifdef __GNUC__
__inline
#endif
int _sock_spill(SOCK *s)
{
  int i0;
  int fd;
  int t;
  register FIFOBUF *fib;

  if (!s->fib_out->enqueued) return(0);
  
  fd = SOCK_FDWRITE(s);
  fib = s->fib_out;
  i0 = fib->bufsize - fib->byte_out;

  if (!i0)
  {
    fib->byte_out = 0;
    fib->node_out = fib->node_out->next;
    i0 = fib->bufsize;
  }

  fcntl(fd, F_SETFL, O_NONBLOCK);

  t = write(fd, fib->node_out->data + fib->byte_out,
            fib->enqueued < i0 ? fib->enqueued : i0);
  
  if (t < 0) _sock_chkerrno(s);
  else
  {
    fib->byte_out += t;
    fib->enqueued -= t;
  }

  fcntl(fd, F_SETFL, 0);
  return(t);  /* For _sock_feed() symmetry. */
}


int sock_io_block(SOCK *s)
{
  fd_set socks_read, socks_write;
  int t;

  FD_ZERO(&socks_read); FD_ZERO(&socks_write);
  FD_SET(SOCK_FDREAD(s), &socks_read);
  if (sock_buf_out_size(s)) FD_SET(SOCK_FDWRITE(s), &socks_write);

  t = select(MAX(SOCK_FDREAD(s), SOCK_FDWRITE(s)) + 1, &socks_read, &socks_write, 0, 0);
  if (t < 1) return(-1);  /* Could be an error condition */

  if (FD_ISSET(SOCK_FDREAD(s), &socks_read))
    if (!(t = _sock_feed(s))) return(-1);

  if (FD_ISSET(SOCK_FDWRITE(s), &socks_write)) _sock_spill(s);
  
  return(0);
}


/* ---- *
 * Read *
 * ---- */


int sock_read(SOCK *s, void *dest, unsigned int size)
{
  int fd, t;
  unsigned int dequeued;

  if (s->cip && s->cip_state != SOCK_CIP_STATE_DONE) _sock_feed(s);
  else
  {
    t = fifobuf_dequeue(s->fib_in, dest, size);
    dest += t;
    size -= t;
  }

  if (size)
  {
    if (!s->cip && !s->fib_in->cip_de)
    {
      /* Saves copying, since we don't have to decrypt */
      
      for (fd = SOCK_FDREAD(s); size; )
      {
        fcntl(fd, F_SETFL, 0);
        t = read(fd, dest, size);
        if (t < 1 && errno != EINTR) break;
        dest += t;
        size -= t;
      }
    }
    else
    {
      /* We have to decrypt, so run it through FIFO */
      
      for (fd = SOCK_FDREAD(s); size; )
      {    
        /* Block */
        
        if (sock_io_block(s) < 0) break;

        dequeued = fifobuf_dequeue(s->fib_in, dest, size);
        size -= dequeued;
        dest += dequeued;
      }
    }

    if (size) { _sock_chkerrno(s); return(-1); }
  }
  
  return(0);
}


void *sock_dread(SOCK *s, unsigned int size)
{
  void *dest;
  
  dest = malloc(size);
  if (!sock_read(s, dest, size)) return(dest);

  free(dest);
  return(0);
}


int sock_getc(SOCK *s)
{
  unsigned char c;

#ifdef TRIES
  if (!s->fib_in->enqueued) _sock_feed(s);
#endif
  if (!sock_read(s, &c, sizeof(unsigned char))) return((int) c);
  return(-1);
}


int sock_gets(SOCK *s, char *dest, unsigned int size_max)
{
  int c;
  unsigned int size;

  size = 0;
  while (size < size_max && (c = sock_getc(s)) != '\n')
  {
    if (c < 0) { dest[size] = 0; return(-1); }
    if (c == '\r') continue;
    dest[size++] = c;
  }

  dest[size] = 0;
  return(size);
}


int _sock_fifobuf_lf(void *data, unsigned long len, void *extra)
{
  unsigned char *p;

  p = memchr(data, '\n', len);

  if (p)
  {
    (*((unsigned int *) extra)) += (p - (unsigned char *) data);
    return(-1);
  }

  (*((unsigned int *) extra)) += len;
  return(0);
}


char *sock_dgets(SOCK *s)
{
  char *r = 0, *p0;
  unsigned int lf;
  int fd, t;
  fd_set sock_read;
  char buf[256];
  unsigned int len, len_alloc;


  if (!s->fib_in->cip_de)
  {
    /* FIFO is not encrypted, so we can operate directly on it and thus
     * save some time. */
    
    lf = 0;
    if (fifobuf_do(s->fib_in, s->fib_in->enqueued, _sock_fifobuf_lf, (void *) &lf))
    {
      r = fifobuf_dequeue_dyn(s->fib_in, (unsigned long) lf + 1);
      *(r + lf) = 0;
      if (lf && *(r + lf - 1) == '\r') *(r + lf - 1) = 0;
      if (fifobuf_peek(s->fib_in, buf, 1) && buf[0] == '\r')
        fifobuf_drop(s->fib_in, 1);
      return(r);
    }

    fd = SOCK_FDREAD(s);
    fcntl(fd, F_SETFL, O_NONBLOCK);

    for (;;)
    {
#if 0    
      printf("[sock_dgets blocking]"); fflush(stdout);
#endif
      FD_ZERO(&sock_read);
      FD_SET(fd, &sock_read);
      t = select(fd + 1, &sock_read, 0, 0, 0);
      if (t < 0 && !_sock_chkerrno(s)) break;
      
      t = read(fd, buf, 255);
      if (t < 1)
      {
#if 0      
        printf("[sock_dgets got empty select]"); fflush(stdout);
#endif
        if (!_sock_chkerrno(s)) break;
        continue;
      }
      
      p0 = memchr(buf, '\n', t);
      if (p0)
      {
        int len;
        
        len = p0 - buf;
        
        if (len && *(p0 - 1) == '\r') len--;
        r = malloc(s->fib_in->enqueued + len + 1);
        memcpy(r + s->fib_in->enqueued, buf, len);
        *((r + s->fib_in->enqueued) + len) = 0;
        fifobuf_dequeue(s->fib_in, r, s->fib_in->enqueued);
        p0++; /* Skip newline */
        
        if (p0 < (buf + t))
        {
          if (*p0 == '\r') p0++;
          if (p0 < (buf + t))
          {
#if 0
            printf("[sock_dgets leftover: %d bytes]\n\n", t - (p0 - buf));
            fwrite(p0 + 1, t - (p0 - buf), 1, stdout);
#endif
            fifobuf_enqueue(s->fib_in, p0 + 1, t - (p0 + 1 - buf));
          }
        }
        
        break;
      }
      
      fifobuf_enqueue(s->fib_in, buf, t);
#if 0
      printf("[sock_dgets got %d bytes]", t); fflush(stdout);
#endif
    }

    fcntl(fd, F_SETFL, 0);
  }
  else
  {
    /* FIFO is encrypted; do this the hard way */
    
    for (len_alloc = 0, len = 0, p0 = 0; ; )
    {
      /* Get byte */
      t = sock_getc(s);
      
      /* Did it work? */
      if (t < 0)
      {
        if (len_alloc) free(p0);
        r = 0; break;
      }
      
      /* End of line? */
      if (t == '\n') 
      {
        p0 = realloc(p0, len + 1);
        *(p0 + len) = 0;
        r = p0; break;
      }
      /* Trailing \r from last line, or \r\n sequence? (Discard) */
      else if (t == '\r') continue;

      /* Move on, expand buffer if required */
      if (len >= len_alloc) p0 = realloc(p0, len_alloc + 1024);
      
      *(p0 + len++) = t;
    }
  }
  
  return(r);
}


/* FIXME: Some systems wholly lack v*scanf. Compensating for this
 *        involves a bit of work. */

#ifdef HAVE_VSSCANF

int sock_scanf(SOCK *s, char *format, ...)
{
  va_list args;
  char *str;
  int ret;

  if (!(str = sock_dgets(s))) return (0);

  va_start(args, format);
  ret = vsscanf(str, format, args);
  va_end(args);
  free(str);
  
  return(ret);
}

#endif


/* ----- *
 * Write *
 * ----- */

int _sock_write(void *data, unsigned long len, void *s)
{
  int t;
  int fd;

  fd = SOCK_FDWRITE((SOCK *) s);

  while (len)
  {
    t = write(fd, data, len);
    if (t < 0 && errno != EINTR) return(-1);
    data += t;
    len -= t;
  }

  return(0);
}


int sock_flush(SOCK *s)
{
  int n;

  if (!(s->flags & SOCK_CONNECTED)) return(-1);

  
  n = fifobuf_do(s->fib_out, s->fib_out->enqueued, _sock_write, s);
  if (n > 0)
  {
    fifobuf_drop(s->fib_out, n);
    _sock_chkerrno(s);
    return(-1);
  }

  fifobuf_drop(s->fib_out, s->fib_out->enqueued);
  return(0);
}


int sock_write(SOCK *s, void *data, unsigned int size)
{
  unsigned int queued;

  if (s->cip)
  {
    /* Finish key exchange first */

    while (s->cip_state != SOCK_CIP_STATE_DONE)
      if (sock_io_block(s) < 0) return(-1);
  }
  
  /* Data too large for max buffer? */
  
  while (size > fifobuf_free_cur(s->fib_out))
  {
    queued = fifobuf_free_cur(s->fib_out);
    
    fifobuf_enqueue(s->fib_out, data, fifobuf_free_cur(s->fib_out));
    size -= queued;
    data += queued;

    if (sock_flush(s) < 0) return(-1);
  }

  /* Remainder */
  
  if (size)
  {
    fifobuf_enqueue(s->fib_out, data, size);
#ifdef TRIES
    _sock_spill(s);
#endif
  }

  return(0);
#if 0
  return(_sock_write(data, size, s));
#endif
}


int sock_putc(SOCK *s, char c)
{
  if (!fifobuf_free_cur(s->fib_out))
    if (sock_flush(s) < 0) return(-1);

  fifobuf_enqueue(s->fib_out, &c, sizeof(char));
#ifdef TRIES
  if (s->fib_out->enqueued > 16) _sock_spill(s);
#endif
  return(0);
}


int sock_puts(SOCK *s, char *str)
{
  int len;
  
  len = strlen(str);
  if (sock_write(s, str, len) < 0) return(-1);

  return(len);
}


int sock_printf(SOCK *s, char *format, ...)
{
  int len;
  va_list args;
  char *str = 0;
  
  va_start(args, format);
  len = vasprintf(&str, format, args);
  va_end(args);
  
  if (len < 1)
  {
    if (str) free(str);
    return(-1);
  }
  
  if (sock_write(s, str, len) < 0) 
  {
    free(str);
    return(-1);
  }

  free(str);
  return(len);
}


int sock_vprintf(SOCK *s, char *format, va_list args)
{
  int len;
  char *str = 0;
  
  len = vasprintf(&str, format, args);
  
  if (len < 1)
  {
    if (str) free(str);
    return(-1);
  }
  
  if (sock_write(s, str, len) < 0) 
  {
    free(str);
    return(-1);
  }

  free(str);
  return(len);
}


/* ------------------ *
 * Cipher negotiation *
 * ------------------ */

void sock_cipher_pull(SOCK *s)
{
  byte *buf;
  unsigned int len;
  
  if (s->cip_state == SOCK_CIP_STATE_GET_ASYM)
  {
    /* We're getting the public part of an asymmetric key */
    
    if (s->cip_buf_ndone < 3)
    {
      if (!s->cip_buf) s->cip_buf = malloc(3);
      s->cip_buf_ndone += fifobuf_dequeue(s->fib_in, s->cip_buf + s->cip_buf_ndone, 3 - s->cip_buf_ndone);
      if (s->cip_buf_ndone == 3) s->cip_buf = realloc(s->cip_buf, 3 + (s->cip_buf[1] << 8) + s->cip_buf[2]);
    }
    
    if (s->cip_buf_ndone >= 3)
    {
      s->cip_buf_ndone +=
        fifobuf_dequeue(s->fib_in, s->cip_buf + s->cip_buf_ndone,
                        (s->cip_buf[1] << 8) + s->cip_buf[2] -
                        s->cip_buf_ndone + 3);
      
      if (s->cip_buf_ndone == (s->cip_buf[1] << 8) + s->cip_buf[2] + 3)
      {
        /* Finished getting, now import it */
        
        cipher_import_asym(s->cip, s->cip_buf);
        free(s->cip_buf);
        
        /* Export, encrypt and enqueue our symmetric key */
        
        cipher_export_sym(s->cip, &buf, &len);
        
        /* if (s->cip->asym_type == CIPHER_ELGAMAL) */
        elgamal_encrypt(s->cip->asym_key_encrypt, (void **) &s->cip_buf, (size_t *) &s->cip_buf_ndone,
                        buf, len);
        free(buf);
        
        fifobuf_enqueue(s->fib_out, s->cip_buf, s->cip_buf_ndone);

        if (s->flags & SOCK_DEBUG)
          log_put_opt(LOG_DEBUG, 0, "[Sock] Sent encrypted key - length is %d.", s->cip_buf_ndone);
        free(s->cip_buf);
        
        /* Permute the _decryption_ key and enable buffer ciphering */
        
        cipher_permute_in(s->cip);
        fifobuf_set_encrypt(s->fib_out, s->cip);
        fifobuf_set_decrypt(s->fib_in, s->cip);
        
        /* Reset cipher metadata */
        
        s->cip_buf = 0;
        s->cip_buf_ndone = 0;
        s->cip_state = SOCK_CIP_STATE_DONE;
      }
    }  /* s->cip_buf_ndone >= 3 */
  }
  else if (s->cip_state == SOCK_CIP_STATE_GET_SYM)
  {
    /* We're getting a symmetric key, encrypted with ElGamal */

    if (!s->cip_buf) s->cip_buf = malloc(elgamal_block_size(s->cip->asym_key_decrypt));

    s->cip_buf_ndone +=
      fifobuf_dequeue(s->fib_in, s->cip_buf + s->cip_buf_ndone,
                      elgamal_block_size(s->cip->asym_key_decrypt) -
                      s->cip_buf_ndone);

    if (s->cip_buf_ndone == elgamal_block_size(s->cip->asym_key_decrypt))
    {
      /* Finished getting, now import it */
      
      elgamal_decrypt(s->cip->asym_key_decrypt, (void **) &buf, &len,
                      s->cip_buf, s->cip_buf_ndone);
      free(s->cip_buf);

      if (s->flags & SOCK_DEBUG)
        log_put_opt(LOG_DEBUG, 0, "[Sock] Got encrypted key - length is %d -> %d.", s->cip_buf_ndone, len);

      cipher_import_sym(s->cip, buf);
      free(buf);

      /* Permute the _encryption_ key and enable buffer ciphering */
      
      cipher_permute_out(s->cip);
      fifobuf_set_encrypt(s->fib_out, s->cip);
      fifobuf_set_decrypt(s->fib_in, s->cip);
      
      /* Reset cipher metadata */
      
      s->cip_buf = 0;
      s->cip_buf_ndone = 0;
      s->cip_state = SOCK_CIP_STATE_DONE;
    }
  }

  if (s->flags & SOCK_DEBUG)
  {
    if (s->cip_buf_ndone >= 3)
    {
      if (s->cip_state == SOCK_CIP_STATE_GET_ASYM)
        log_put_opt(LOG_DEBUG, 0, "[Sock] Got cipher data - ndone is %d/%d.",
                    s->cip_buf_ndone, (s->cip_buf[1] << 8) + s->cip_buf[2]);
      else
        log_put_opt(LOG_DEBUG, 0, "[Sock] Got cipher data - ndone is %d/%d.",
                    s->cip_buf_ndone, s->cip_buf[1] + s->cip_buf[2]);
    }
    else
      log_put_opt(LOG_DEBUG, 0, "[Sock] Got cipher data - ndone is %d.", s->cip_buf_ndone);
  }
}


void sock_cipher_exchange(SOCK *s)
{
  if (!s->cip) return;

  if (s->flags & SOCK_DEBUG)
    log_put_opt(LOG_DEBUG, 0, "[Sock] Starting blocking key exchange.");

  while (s->cip_state != SOCK_CIP_STATE_DONE) sock_io_block(s);

  if (s->flags & SOCK_DEBUG)
    log_put_opt(LOG_DEBUG, 0, "[Sock] Finished blocking key exchange.");
}
