/* Copyright (C) 1999-2000 Hans Petter K. Jansson
 *
 * This library is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
 *
 * You can contact the library's author by sending e-mail to <hpj@styx.net>.
 */

#include "config.h"
#include "test.h"
#include "flux.h"
#include "cipher.h"
#include <stdlib.h>
#include <sys/time.h>
#include <sys/resource.h>
#include <unistd.h>


CIPHER_ELGAMAL_SKEY *skey;  /* Carried over between tests */


/* More complete test program. This does 1000 encryptions and decryptions
 * with each of 250 128-bit keys and 2000 encryptions and decryptions with
 * each of 125 256-bit keys, using a feedback scheme similar to a Feistel
 * cipher, so as to be sure of testing all the table entries pretty
 * thoroughly.  We keep changing the keys so as to get a more meaningful
 * performance number, since the key setup is non-trivial for Twofish. */

int test_cipher_twofish()
{
  CIPHER_TWOFISH_KEY ctx;     /* Expanded key. */
  int i, j;    /* Loop counters. */
  int ret = 0;

  /* Test buffer. */
  byte buffer[4][16] =
  {
    {0x00, 0x11, 0x22, 0x33, 0x44, 0x55, 0x66, 0x77,
     0x88, 0x99, 0xAA, 0xBB, 0xCC, 0xDD, 0xEE, 0xFF},
    {0x0F, 0x1E, 0x2D, 0x3C, 0x4B, 0x5A, 0x69, 0x78,
     0x87, 0x96, 0xA5, 0xB4, 0xC3, 0xD2 ,0xE1, 0xF0},
    {0x01, 0x23, 0x45, 0x67, 0x89, 0xAB, 0xCD, 0xEF,
     0xFE, 0xDC, 0xBA, 0x98, 0x76, 0x54 ,0x32, 0x10},
    {0x01, 0x23, 0x45, 0x67, 0x76, 0x54 ,0x32, 0x10,
     0x89, 0xAB, 0xCD, 0xEF, 0xFE, 0xDC, 0xBA, 0x98}
  };

  /* Expected outputs for the million-operation test */
  static const byte test_encrypt[4][16] =
  {
    {0xC8, 0x23, 0xB8, 0xB7, 0x6B, 0xFE, 0x91, 0x13,
     0x2F, 0xA7, 0x5E, 0xE6, 0x94, 0x77, 0x6F, 0x6B},
    {0x90, 0x36, 0xD8, 0x29, 0xD5, 0x96, 0xC2, 0x8E,
     0xE4, 0xFF, 0x76, 0xBC, 0xE5, 0x77, 0x88, 0x27},
    {0xB8, 0x78, 0x69, 0xAF, 0x42, 0x8B, 0x48, 0x64,
     0xF7, 0xE9, 0xF3, 0x9C, 0x42, 0x18, 0x7B, 0x73},
    {0x7A, 0x88, 0xFB, 0xEB, 0x90, 0xA4, 0xB4, 0xA8,
     0x43, 0xA3, 0x1D, 0xF1, 0x26, 0xC4, 0x53, 0x57}
  };
  static const byte test_decrypt[4][16] =
  {
    {0x00, 0x11, 0x22, 0x33, 0x44, 0x55, 0x66, 0x77,
     0x88, 0x99, 0xAA, 0xBB, 0xCC, 0xDD, 0xEE, 0xFF},
    {0x0F, 0x1E, 0x2D, 0x3C, 0x4B, 0x5A, 0x69, 0x78,
     0x87, 0x96, 0xA5, 0xB4, 0xC3, 0xD2 ,0xE1, 0xF0},
    {0x01, 0x23, 0x45, 0x67, 0x89, 0xAB, 0xCD, 0xEF,
     0xFE, 0xDC, 0xBA, 0x98, 0x76, 0x54 ,0x32, 0x10},
    {0x01, 0x23, 0x45, 0x67, 0x76, 0x54 ,0x32, 0x10,
     0x89, 0xAB, 0xCD, 0xEF, 0xFE, 0xDC, 0xBA, 0x98}
  };

  bm_start();

  /* Encryption test. */
  for (i = 0; i < 125; i++)
  {
    twofish_set (&ctx, buffer[0], sizeof (buffer[0]));
    for (j = 0; j < 1000; j++) twofish_encrypt(&ctx, buffer[2], buffer[2]);

    twofish_set (&ctx, buffer[1], sizeof (buffer[1]));
    for (j = 0; j < 1000; j++) twofish_encrypt(&ctx, buffer[3], buffer[3]);

    twofish_set (&ctx, buffer[2], sizeof (buffer[2])*2);
    for (j = 0; j < 1000; j++)
    {
      twofish_encrypt(&ctx, buffer[0], buffer[0]);
      twofish_encrypt(&ctx, buffer[1], buffer[1]);
    }
  }
  
  
  if (memcmp(buffer, test_encrypt, sizeof(test_encrypt))) ret = -1;

  /* Decryption test. */
  for (i = 0; i < 125; i++)
  {
    twofish_set (&ctx, buffer[2], sizeof (buffer[2])*2);
    for (j = 0; j < 1000; j++)
    {
      twofish_decrypt(&ctx, buffer[0], buffer[0]);
      twofish_decrypt(&ctx, buffer[1], buffer[1]);
    }
    
    twofish_set(&ctx, buffer[1], sizeof (buffer[1]));
    for (j = 0; j < 1000; j++) twofish_decrypt(&ctx, buffer[3], buffer[3]);

    twofish_set (&ctx, buffer[0], sizeof (buffer[0]));
    for (j = 0; j < 1000; j++) twofish_decrypt(&ctx, buffer[2], buffer[2]);
  }

  if (memcmp(buffer, test_decrypt, sizeof(test_decrypt))) ret = -1;
  
  bm_end();
  return((int) ((double) (250.0 * 1000.0 * 2.0 + 250.0 * 2000.0 * 2.0) * 16.0 * 1000 /
                bm_diff_get_usec()));
}


int test_cipher_twofish_cbc()
{
  CIPHER_TWOFISH_KEY *ctx_enc, *ctx_dec;     /* Expanded key. */
  int i;    /* Loop counter. */

  /* Test buffers. */
  
  byte in[16] =
  {
     0x00, 0x11, 0x22, 0x33, 0x44, 0x55, 0x66, 0x77,
     0x88, 0x99, 0xAA, 0xBB, 0xCC, 0xDD, 0xEE, 0xFF
  };
  
  byte out[16];

  /* Exchange key and IV */
  
  ctx_enc = twofish_gen(128);
  ctx_dec = twofish_dup(ctx_enc);

  if (!ctx_enc) printf("No enc key!\n");
  if (!ctx_dec) printf("No dec key!\n");
  if (memcmp(ctx_enc->k, ctx_dec->k, 16)) printf("Keys don't match!\n");
  if (memcmp(ctx_enc->v.v, ctx_dec->v.v, 16)) printf("IVs don't match!\n");
  
  bm_start();
  
  for (i = 0; i < 200000; i++)
  {
    twofish_encrypt_cbc(ctx_enc, out, in);
    twofish_decrypt_cbc(ctx_dec, out, out);
    if (memcmp(in, out, 16)) return(-1);
  }

  bm_end();
  
  return((int) ((double) 200.0 * 2.0 * 16.0 * 1000000.0 / bm_diff_get_usec()));
}


int test_cipher_twofish_cfb()
{
  CIPHER_TWOFISH_KEY *ctx_enc, *ctx_dec;     /* Expanded key. */
  int i;    /* Loop counter. */

  /* Test buffers. */
  
  byte in[256] =
  {
     0x00, 0x11, 0x22, 0x33, 0x44, 0x55, 0x66, 0x77,
     0x88, 0x99, 0xAA, 0xBB, 0xCC, 0xDD, 0xEE, 0xFF,

     0x00, 0x11, 0x22, 0x33, 0x44, 0x55, 0x66, 0x77,
     0x88, 0x99, 0xAA, 0xBB, 0xCC, 0xDD, 0xEE, 0xFF,

     0x00, 0x11, 0x22, 0x33, 0x44, 0x55, 0x66, 0x77,
     0x88, 0x99, 0xAA, 0xBB, 0xCC, 0xDD, 0xEE, 0xFF,
     
     0x00, 0x11, 0x22, 0x33, 0x44, 0x55, 0x66, 0x77,
     0x88, 0x99, 0xAA, 0xBB, 0xCC, 0xDD, 0xEE, 0xFF,
     
     0x00, 0x11, 0x22, 0x33, 0x44, 0x55, 0x66, 0x77,
     0x88, 0x99, 0xAA, 0xBB, 0xCC, 0xDD, 0xEE, 0xFF,
     
     0x00, 0x11, 0x22, 0x33, 0x44, 0x55, 0x66, 0x77,
     0x88, 0x99, 0xAA, 0xBB, 0xCC, 0xDD, 0xEE, 0xFF,
     
     0x00, 0x11, 0x22, 0x33, 0x44, 0x55, 0x66, 0x77,
     0x88, 0x99, 0xAA, 0xBB, 0xCC, 0xDD, 0xEE, 0xFF,
     
     0x00, 0x11, 0x22, 0x33, 0x44, 0x55, 0x66, 0x77,
     0x88, 0x99, 0xAA, 0xBB, 0xCC, 0xDD, 0xEE, 0xFF,
     
     0x00, 0x11, 0x22, 0x33, 0x44, 0x55, 0x66, 0x77,
     0x88, 0x99, 0xAA, 0xBB, 0xCC, 0xDD, 0xEE, 0xFF,
     
     0x00, 0x11, 0x22, 0x33, 0x44, 0x55, 0x66, 0x77,
     0x88, 0x99, 0xAA, 0xBB, 0xCC, 0xDD, 0xEE, 0xFF,
    
     0x00, 0x11, 0x22, 0x33, 0x44, 0x55, 0x66, 0x77,
     0x88, 0x99, 0xAA, 0xBB, 0xCC, 0xDD, 0xEE, 0xFF,
     
     0x00, 0x11, 0x22, 0x33, 0x44, 0x55, 0x66, 0x77,
     0x88, 0x99, 0xAA, 0xBB, 0xCC, 0xDD, 0xEE, 0xFF,
     
     0x00, 0x11, 0x22, 0x33, 0x44, 0x55, 0x66, 0x77,
     0x88, 0x99, 0xAA, 0xBB, 0xCC, 0xDD, 0xEE, 0xFF,
     
     0x00, 0x11, 0x22, 0x33, 0x44, 0x55, 0x66, 0x77,
     0x88, 0x99, 0xAA, 0xBB, 0xCC, 0xDD, 0xEE, 0xFF,
     
     0x00, 0x11, 0x22, 0x33, 0x44, 0x55, 0x66, 0x77,
     0x88, 0x99, 0xAA, 0xBB, 0xCC, 0xDD, 0xEE, 0xFF,
     
     0x00, 0x11, 0x22, 0x33, 0x44, 0x55, 0x66, 0x77,
     0x88, 0x99, 0xAA, 0xBB, 0xCC, 0xDD, 0xEE, 0xFF
  };
  
  byte out[256];

  /* Exchange key and IV */
  
  ctx_enc = twofish_gen(128);
  ctx_dec = twofish_dup(ctx_enc);

  if (!ctx_enc) printf("No enc key!\n");
  if (!ctx_dec) printf("No dec key!\n");
  if (memcmp(ctx_enc->k, ctx_dec->k, 16)) printf("Keys don't match!\n");
  if (memcmp(ctx_enc->v.v, ctx_dec->v.v, 16)) printf("IVs don't match!\n");
  
  bm_start();
  
  for (i = 0; i < 500; i++)
  {
    twofish_encrypt_cfb(ctx_enc, out, in, 256);
    twofish_decrypt_cfb(ctx_dec, out, out, 256);
  }

  if (memcmp(in, out, 256)) return(-1);

  bm_end();

  return((int) ((double) 0.5 * 2.0 * 256.0 * 1000000.0 / bm_diff_get_usec()));
}


int test_cipher_twofish_ofb()
{
  CIPHER_TWOFISH_KEY *ctx_enc, *ctx_dec;     /* Expanded key. */
  int i;    /* Loop counter. */

  /* Test buffers. */
  
  byte in[256] =
  {
     0x00, 0x11, 0x22, 0x33, 0x44, 0x55, 0x66, 0x77,
     0x88, 0x99, 0xAA, 0xBB, 0xCC, 0xDD, 0xEE, 0xFF,

     0x00, 0x11, 0x22, 0x33, 0x44, 0x55, 0x66, 0x77,
     0x88, 0x99, 0xAA, 0xBB, 0xCC, 0xDD, 0xEE, 0xFF,

     0x00, 0x11, 0x22, 0x33, 0x44, 0x55, 0x66, 0x77,
     0x88, 0x99, 0xAA, 0xBB, 0xCC, 0xDD, 0xEE, 0xFF,
     
     0x00, 0x11, 0x22, 0x33, 0x44, 0x55, 0x66, 0x77,
     0x88, 0x99, 0xAA, 0xBB, 0xCC, 0xDD, 0xEE, 0xFF,
     
     0x00, 0x11, 0x22, 0x33, 0x44, 0x55, 0x66, 0x77,
     0x88, 0x99, 0xAA, 0xBB, 0xCC, 0xDD, 0xEE, 0xFF,
     
     0x00, 0x11, 0x22, 0x33, 0x44, 0x55, 0x66, 0x77,
     0x88, 0x99, 0xAA, 0xBB, 0xCC, 0xDD, 0xEE, 0xFF,
     
     0x00, 0x11, 0x22, 0x33, 0x44, 0x55, 0x66, 0x77,
     0x88, 0x99, 0xAA, 0xBB, 0xCC, 0xDD, 0xEE, 0xFF,
     
     0x00, 0x11, 0x22, 0x33, 0x44, 0x55, 0x66, 0x77,
     0x88, 0x99, 0xAA, 0xBB, 0xCC, 0xDD, 0xEE, 0xFF,
     
     0x00, 0x11, 0x22, 0x33, 0x44, 0x55, 0x66, 0x77,
     0x88, 0x99, 0xAA, 0xBB, 0xCC, 0xDD, 0xEE, 0xFF,
     
     0x00, 0x11, 0x22, 0x33, 0x44, 0x55, 0x66, 0x77,
     0x88, 0x99, 0xAA, 0xBB, 0xCC, 0xDD, 0xEE, 0xFF,
    
     0x00, 0x11, 0x22, 0x33, 0x44, 0x55, 0x66, 0x77,
     0x88, 0x99, 0xAA, 0xBB, 0xCC, 0xDD, 0xEE, 0xFF,
     
     0x00, 0x11, 0x22, 0x33, 0x44, 0x55, 0x66, 0x77,
     0x88, 0x99, 0xAA, 0xBB, 0xCC, 0xDD, 0xEE, 0xFF,
     
     0x00, 0x11, 0x22, 0x33, 0x44, 0x55, 0x66, 0x77,
     0x88, 0x99, 0xAA, 0xBB, 0xCC, 0xDD, 0xEE, 0xFF,
     
     0x00, 0x11, 0x22, 0x33, 0x44, 0x55, 0x66, 0x77,
     0x88, 0x99, 0xAA, 0xBB, 0xCC, 0xDD, 0xEE, 0xFF,
     
     0x00, 0x11, 0x22, 0x33, 0x44, 0x55, 0x66, 0x77,
     0x88, 0x99, 0xAA, 0xBB, 0xCC, 0xDD, 0xEE, 0xFF,
     
     0x00, 0x11, 0x22, 0x33, 0x44, 0x55, 0x66, 0x77,
     0x88, 0x99, 0xAA, 0xBB, 0xCC, 0xDD, 0xEE, 0xFF
  };
  
  byte out[256];

  /* Exchange key and IV */
  
  ctx_enc = twofish_gen(128);
  ctx_dec = twofish_dup(ctx_enc);

  if (!ctx_enc) printf("No enc key!\n");
  if (!ctx_dec) printf("No dec key!\n");
  if (memcmp(ctx_enc->k, ctx_dec->k, 16)) printf("Keys don't match!\n");
  if (memcmp(ctx_enc->v.v, ctx_dec->v.v, 16)) printf("IVs don't match!\n");
  
  bm_start();
  
  for (i = 0; i < 5000; i++)
  {
    twofish_encrypt_ofb(ctx_enc, out, in, 256);
    twofish_decrypt_ofb(ctx_dec, out, out, 256);
  }

  if (memcmp(in, out, 256)) return(-1);

  bm_end();

  return((int) ((double) 5.0 * 2.0 * 256.0 * 1000000.0 / bm_diff_get_usec()));
}


int test_cipher_elgamal()
{
  CIPHER_ELGAMAL_PKEY *pkey0, *pkey1;
  void *ciphertext, *plaintext;
  size_t ciphertext_len, plaintext_len;
  byte *buf;
  unsigned int buf_len;

  /* Key generation */

/*  
  printf("[key] "); fflush(stdout);
*/

  skey = elgamal_gen_secret(768);
  if (!skey) return(-1);

  pkey0 = elgamal_gen_public(skey);
  elgamal_export_public(pkey0, &buf, &buf_len);
  pkey1 = elgamal_import_public(buf);
  free(buf);

  if (!elgamal_is_pair(pkey1, skey)) return(-1);

  /* Encrypt/decrypt */

/*
  printf("[encrypt] "); fflush(stdout);
*/

  plaintext_len = 22;
  elgamal_encrypt(pkey1, &ciphertext, &ciphertext_len,
                  "Dark Side Energy Bolts", plaintext_len);

/*  
  printf("[decrypt] "); fflush(stdout);
*/
  
  elgamal_decrypt(skey, &plaintext, &plaintext_len,
                  ciphertext, ciphertext_len);

  free(ciphertext);
  
/*
  fwrite(plaintext, plaintext_len, 1, stdout);
  fputc('\n', stdout);
  fflush(stdout);
*/
  
  if (memcmp(plaintext, "Dark Side Energy Bolts", plaintext_len)) return(-1);
  
  free(plaintext);

  /* Sign/verify */

/*
  printf("[sign] "); fflush(stdout);
*/
  
  elgamal_sign(skey, &ciphertext, &ciphertext_len,
               "On wednesdays I go shopping, and have buttered scones for tea.",
               62);
/*  
  printf("[verify] "); fflush(stdout);
*/
  
  if (!elgamal_verify(pkey1, "On wednesdays I go shopping, and have buttered scones for tea.",
                      62, ciphertext, ciphertext_len))
    return(-1);
  
  free(ciphertext);
  return(1);
}
