/* Copyright (C) 1999-2000 Hans Petter K. Jansson
 *
 * This library is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
 *
 * You can contact the library's author by sending e-mail to <hpj@styx.net>.
 */

#include "config.h"
#include "test.h"
#include "bt.h"
#include "cipher.h"
#include "flux.h"
#include <stdlib.h>
#include <unistd.h>


extern CIPHER_ELGAMAL_SKEY *skey;


int test_sock_connect()
{
  SOCK *s0, *s1, *s2;
  
  s0 = sock_accepting_new(2596);
  if (!s0) return(-1);
  s1 = sock_new();
  if (!s1)
  {
    sock_del(s0);
    return(-1);
  }
  
  sock_connect(s1, "127.0.0.1", 2596);
  s2 = sock_accept(s0);
  if (!s2)
  {
    sock_del(s0);
    sock_del(s1);
    return(-1);
  }

  sock_del(s2);
  sock_del(s0);
  sock_del(s1);
  return(0);
}


int test_sock_transfer()
{
  SOCK *s0, *s1, *s2;
  char str[128];
  
  s0 = sock_accepting_new(2597);
  if (!s0) return(-1);

  if (!fork())
  {
    /* Child */

    s1 = sock_new();
    if (!s1)
    {
      sock_del(s0);
      exit(2);
    }

    sock_connect(s1, "127.0.0.1", 2597);
    sock_puts(s1, "Your mother.");
    sock_flush(s1);

    exit(0);
  }

  /* Parent */

  s2 = sock_wait(0, 0, s0, 0);
  if (!s2)
  {
    sock_del(s0);
    return(-1);
  }

  sock_gets(s2, str, 128);

  if (!str || strcmp(str, "Your mother."))
  {
    printf("[Got \"%s\"] ", str);
    sock_del(s0);
    sock_del(s2);
    return(-1);
  }

  sock_del(s0);
  sock_del(s2);
  return(0);
}


int test_sock_crypto()
{
  CIPHER_TWOFISH_KEY *tf_key_en, *tf_key_de;
  CIPHER *cip0, *cip1;
  SOCK *s0, *s1, *s2;
  char str[128];

  tf_key_en = twofish_gen(128);
  tf_key_de = twofish_dup(tf_key_en);

  cip0 = malloc(sizeof(*cip0));
  cip0->sym_key_encrypt = tf_key_en;
  cip0->sym_key_decrypt = tf_key_de;

  s0 = sock_accepting_new(2598);
  if (!s0) return(-1);

  if (!fork())
  {
    /* Child */

    s1 = sock_new();
    if (!s1) exit(2);

    cip1 = malloc(sizeof(*cip1));
    cip1->sym_key_encrypt = twofish_dup(tf_key_en);
    cip1->sym_key_decrypt = twofish_dup(tf_key_de);
  
    s1->fib_out->cip_en = cip1;
    s1->fib_in->cip_de = cip1;
    
    sock_connect(s1, "127.0.0.1", 2598);
    sock_puts(s1, "Your mother.\r\n");
    sock_flush(s1);
    sock_del(s1);

    exit(0);
  }

  /* Parent */

  s2 = sock_wait(0, 0, s0, 0);
  if (!s2)
  {
    sock_del(s0);
    return(-1);
  }

  s2->fib_out->cip_en = cip0;
  s2->fib_in->cip_de = cip0;

  sock_gets(s2, str, 128);

  if (!str || strcmp(str, "Your mother."))
  {
    printf("[Got \"%s\"] ", str);
    sock_del(s0);
    sock_del(s2);
    return(-1);
  }

  sock_del(s0);
  sock_del(s2);
  return(0);
}


int test_sock_session_crypto()
{
  CIPHER *cip0, *cip1;
  SOCK *s0, *s1, *s2;
  char str[128];

  cip0 = cipher_new(CIPHER_NONE, CIPHER_ELGAMAL);
  cip1 = cipher_new(CIPHER_TWOFISH, CIPHER_NONE);

  cip0->asym_key_encrypt = elgamal_gen_public(skey);
  cip0->asym_key_decrypt = skey;
  cip0->asym_key_sign = skey;
  cip0->asym_key_verify = elgamal_gen_public(skey);
  
  s0 = sock_accepting_new(2599);
  if (!s0) return(-1);
  sock_set_cipher(s0, cip0);

  if (!fork())
  {
    /* Child */

    s1 = sock_new();
    if (!s1)
    {
      sock_del(s1);
      exit(2);
    }

    sock_set_cipher(s1, cip1);
    
    sock_connect(s1, "127.0.0.1", 2599);
    sock_gets(s1, str, 128);
    
    if (!str || strcmp(str, "Elderberries."))
    {
      printf("[Got \"%s\"] ", str);
      sock_del(s1);
      return(-1);
    }

    sock_puts(s1, "Your mother.\r\n");
    sock_flush(s1);

    sock_del(s1);
    exit(0);
  }

  /* Parent */

  s2 = sock_wait(0, 0, s0, 0);
  if (!s2)
  {
    sock_del(s0);
    return(-1);
  }

  sock_puts(s2, "Elderberries.\r\n");
  sock_flush(s2);

  sock_gets(s2, str, 128);

  if (!str || strcmp(str, "Your mother."))
  {
    printf("[Got \"%s\"] ", str);
    sock_del(s0);
    sock_del(s2);
    return(-1);
  }
  
  sock_del(s0);
  sock_del(s2);
  return(0);
}
